# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import io
import os
import random
import string

from typing import List

from rebasehelper.helpers.process_helper import ProcessHelper


class TestProcessHelper:
    """ ProcessHelper tests """

    class TestRunSubprocess:
        """ ProcessHelper - run_subprocess() tests """
        TEMP_FILE: str = "temp_file"
        TEMP_DIR: str = "temp_dir"
        OUT_FILE: str = "output_file"
        IN_FILE: str = "input_file"
        PHRASE: str = "hello world"
        ECHO_COMMAND: List[str] = ["echo", PHRASE]
        TOUCH_COMMAND: List[str] = ["touch", TEMP_FILE]
        LS_COMMAND: List[str] = ["ls"]
        CAT_COMMAND: List[str] = ["cat"]

        def test_simple_cmd(self):
            ret = ProcessHelper.run_subprocess(self.TOUCH_COMMAND)
            assert ret == 0
            assert os.path.exists(self.TEMP_FILE)

        def test_simple_cmd_with_redirected_output_path(self):
            ret = ProcessHelper.run_subprocess(self.ECHO_COMMAND,
                                               output_file=self.OUT_FILE)
            assert ret == 0
            assert os.path.exists(self.OUT_FILE)
            assert open(self.OUT_FILE).readline().strip('\n') == self.PHRASE

        def test_simple_cmd_with_redirected_output_fileobject(self):
            buff = io.StringIO()
            ret = ProcessHelper.run_subprocess(self.ECHO_COMMAND,
                                               output_file=buff)
            assert ret == 0
            assert not os.path.exists(self.OUT_FILE)
            assert buff.readline().strip('\n') == self.PHRASE
            buff.close()

        def test_simple_cmd_with_input_path_and_redirected_output_path(self):
            with open(self.IN_FILE, 'w') as f:
                f.write(self.PHRASE)

            assert os.path.exists(self.IN_FILE)
            assert open(self.IN_FILE).readline().strip('\n') == self.PHRASE

            ret = ProcessHelper.run_subprocess(self.CAT_COMMAND,
                                               input_file=self.IN_FILE,
                                               output_file=self.OUT_FILE)
            assert ret == 0
            assert os.path.exists(self.OUT_FILE)
            assert open(self.OUT_FILE).readline().strip('\n') == self.PHRASE

        def test_simple_cmd_with_input_fileobject_and_redirected_output_path(self):
            in_buff = io.StringIO()
            in_buff.write(self.PHRASE)

            assert not os.path.exists(self.IN_FILE)
            in_buff.seek(0)
            assert in_buff.readline().strip('\n') == self.PHRASE

            ret = ProcessHelper.run_subprocess(self.CAT_COMMAND,
                                               input_file=in_buff,
                                               output_file=self.OUT_FILE)
            in_buff.close()
            assert ret == 0
            assert os.path.exists(self.OUT_FILE)
            assert open(self.OUT_FILE).readline().strip('\n') == self.PHRASE

        def test_simple_cmd_with_input_path_and_redirected_output_fileobject(self):
            out_buff = io.StringIO()
            with open(self.IN_FILE, 'w') as f:
                f.write(self.PHRASE)

            assert os.path.exists(self.IN_FILE)
            assert open(self.IN_FILE).readline().strip('\n') == self.PHRASE

            ret = ProcessHelper.run_subprocess(self.CAT_COMMAND,
                                               input_file=self.IN_FILE,
                                               output_file=out_buff)
            assert ret == 0
            assert not os.path.exists(self.OUT_FILE)
            out_buff.seek(0)
            assert out_buff.readline().strip('\n') == self.PHRASE
            out_buff.close()

        def test_simple_cmd_with_input_fileobject_and_redirected_output_fileobject(self):
            out_buff = io.StringIO()
            in_buff = io.StringIO()
            in_buff.write(self.PHRASE)

            assert not os.path.exists(self.IN_FILE)
            in_buff.seek(0)
            assert in_buff.readline().strip('\n') == self.PHRASE

            ret = ProcessHelper.run_subprocess(self.CAT_COMMAND,
                                               input_file=in_buff,
                                               output_file=out_buff)
            in_buff.close()
            assert ret == 0
            assert not os.path.exists(self.OUT_FILE)
            out_buff.seek(0)
            assert out_buff.readline().strip('\n') == self.PHRASE
            out_buff.close()

    class TestRunSubprocessCwd:
        """ ProcessHelper - run_subprocess_cwd() tests """
        TEMP_FILE: str = "temp_file"
        TEMP_DIR: str = "temp_dir"
        OUT_FILE: str = "output_file"
        PHRASE: str = "hello world"
        ECHO_COMMAND: List[str] = ["echo", PHRASE]
        TOUCH_COMMAND: List[str] = ["touch", TEMP_FILE]
        LS_COMMAND: List[str] = ["ls"]

        def test_simple_cmd_changed_work_dir(self):
            os.mkdir(self.TEMP_DIR)
            ret = ProcessHelper.run_subprocess_cwd(self.TOUCH_COMMAND,
                                                   self.TEMP_DIR)
            assert ret == 0
            assert os.path.exists(os.path.join(self.TEMP_DIR, self.TEMP_FILE))

        def test_simple_cmd_changed_work_dir_with_redirected_output(self):
            # create temp_file in temp_dir
            self.test_simple_cmd_changed_work_dir()
            ret = ProcessHelper.run_subprocess_cwd(self.LS_COMMAND,
                                                   self.TEMP_DIR,
                                                   output_file=self.OUT_FILE)
            assert ret == 0
            assert os.path.exists(os.path.join(self.TEMP_DIR, self.TEMP_FILE))
            assert os.path.exists(self.OUT_FILE)
            assert open(self.OUT_FILE).readline().strip("\n") == self.TEMP_FILE

    class TestRunSubprocessCwdEnv:
        """ ProcessHelper - run_subprocess_cwd_env() tests """
        OUT_FILE: str = "output_file"
        PHRASE: str = "hello world"

        def test_setting_new_env(self):
            # make copy of existing environment
            en_variables = os.environ.copy().keys()

            # pick up non-existing name
            while True:
                rand_name = ''.join(random.choice(string.ascii_letters) for _ in range(6)).upper()
                if rand_name not in en_variables:
                    break

            cmd = 'echo "$' + rand_name + '"'
            ret = ProcessHelper.run_subprocess_cwd_env(cmd,
                                                       env={rand_name: self.PHRASE},
                                                       output_file=self.OUT_FILE,
                                                       shell=True)
            assert ret == 0
            assert os.path.exists(self.OUT_FILE)
            assert open(self.OUT_FILE).readline().strip("\n") == self.PHRASE

        def test_setting_existing_env(self):
            # make copy of existing environment
            en_variables = list(os.environ.copy().keys())

            # there are no variables set on the system -> nothing to test
            if not en_variables:
                pass

            assert os.environ.get(en_variables[0]) != self.PHRASE

            cmd = 'echo "$' + en_variables[0] + '"'
            ret = ProcessHelper.run_subprocess_cwd_env(cmd,
                                                       env={en_variables[0]: self.PHRASE},
                                                       output_file=self.OUT_FILE,
                                                       shell=True)
            assert ret == 0
            assert os.path.exists(self.OUT_FILE)
            assert open(self.OUT_FILE).readline().strip("\n") == self.PHRASE



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

from typing import List

from rebasehelper.plugins.plugin_manager import plugin_manager
from rebasehelper.tests.conftest import TEST_FILES_DIR


class TestPkgDiff:
    FILE_XML: str = "files.xml"
    TEST_FILES: List[str] = [
        FILE_XML,
    ]

    def get_data(self):
        data = {'moved': ['/usr/src/debug/iproute2-3.12.0/test/mytest.c',
                          '/usr/src/debug/iproute2-3.12.0/test/mytest2.c'],
                'changed': ['/usr/include/test.h']}
        return data

    def test_fill_dictionary(self):
        expected_dict = {'added': ['/usr/sbin/test',
                                   '/usr/sbin/pkg-*/binary_test',
                                   '/usr/lib64/libtest.so',
                                   '/usr/lib64/libtest.so.1'],
                         'removed': ['/usr/sbin/my_test',
                                     '/usr/sbin/pkg-*/binary_test',
                                     '/usr/lib64/libtest2.so',
                                     '/usr/lib64/libtest2.so.1'],
                         'changed': ['/usr/share/test.man'],
                         'moved': ['/usr/local/test.sh',
                                   '/usr/sbin/pkg-*/binary_test;/usr/sbin/pkg-*/binary_test (1%)'],
                         'renamed': ['/usr/lib/libtest3.so.3',
                                     '/usr/lib/libtest3.so']}
        plugin_manager.checkers.plugins['pkgdiff'].results_dir = TEST_FILES_DIR
        plugin_manager.checkers.plugins['pkgdiff'].fill_dictionary(TEST_FILES_DIR, old_version='1.0.1',
                                                                   new_version='1.0.2')
        assert plugin_manager.checkers.plugins['pkgdiff'].results_dict == expected_dict

    def test_process_xml(self):
        expected_dict = {'added': ['/usr/sbin/test',
                                   '/usr/lib64/libtest.so',
                                   '/usr/lib64/libtest.so.1'],
                         'changed': ['/usr/share/test.man'],
                         'moved': ['/usr/local/test.sh',
                                   '/usr/sbin/pkg-*/binary_test;/usr/sbin/pkg-*/binary_test (1%)'],
                         'removed': ['/usr/sbin/my_test',
                                     '/usr/lib64/libtest2.so',
                                     '/usr/lib64/libtest2.so.1'],
                         'renamed': ['/usr/lib/libtest3.so.3',
                                     '/usr/lib/libtest3.so']}
        plugin_manager.checkers.plugins['pkgdiff'].results_dir = TEST_FILES_DIR
        res_dict = plugin_manager.checkers.plugins['pkgdiff'].process_xml_results(TEST_FILES_DIR, old_version='1.0.1',
                                                                                  new_version='1.0.2')
        assert res_dict == expected_dict



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os

import git  # type: ignore
import pytest  # type: ignore

from rebasehelper.helpers.git_helper import GitHelper


class TestGitHelper:

    def write_config_file(self, config_file, name, email):
        with open(config_file, 'w') as f:
            f.write('[user]\n'
                    '    name = {0}\n'
                    '    email = {1}\n'.format(name, email))

    @pytest.mark.parametrize('config', [
        # Get from $XDG_CONFIG_HOME/git/config
        'global',
        # Get from included file in $XDG_CONFIG_HOME/git/config
        'global_include',
        # Get from $repo_path/.git/config
        'local',
        # Get from GIT_CONFIG
        'env',
    ])
    def test_get_user_and_email(self, config, workdir):
        name = 'Foo Bar'
        email = 'foo@bar.com'
        env = os.environ.copy()

        try:
            if config == 'global':
                work_git_path = os.path.join(workdir, 'git')
                os.makedirs(work_git_path)

                config_file = os.path.join(work_git_path, 'config')
                self.write_config_file(config_file, name, email)
                os.environ['HOME'] = workdir
                os.environ['XDG_CONFIG_HOME'] = workdir
            elif config == 'global_include':
                work_git_path = os.path.join(workdir, 'git')
                os.makedirs(work_git_path)

                config_file = os.path.join(work_git_path, 'config')
                with open(config_file, 'w') as f:
                    f.write('[include]\n'
                            '    path = included_config\n')
                included_config_file = os.path.join(work_git_path, 'included_config')
                self.write_config_file(included_config_file, name, email)
                os.environ['HOME'] = workdir
                os.environ['XDG_CONFIG_HOME'] = workdir
            elif config == 'local':
                repo = git.Repo.init(workdir)
                repo.git.config('user.name', name, local=True)
                repo.git.config('user.email', email, local=True)
            elif config == 'env':
                config_file = os.path.join(workdir, 'git_config')
                os.environ['GIT_CONFIG'] = config_file
                self.write_config_file(config_file, name, email)
            else:
                raise RuntimeError()

            assert name == GitHelper.get_user()
            assert email == GitHelper.get_email()
        finally:
            os.environ = env



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os

from typing import Optional, cast

from rebasehelper.logger import CustomLogger
from rebasehelper.results_store import results_store
from rebasehelper.helpers.rpm_helper import RpmHelper
from rebasehelper.plugins.checkers import CheckerCategory
from rebasehelper.plugins.checkers.rpminspect import Rpminspect


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class RpminspectSrpm(Rpminspect):
    """SRPM rpminspect checker."""

    CATEGORY: Optional[CheckerCategory] = CheckerCategory.SRPM

    @classmethod
    def run_check(cls, results_dir, **kwargs):
        cls.results_dir = os.path.join(results_dir, 'rpminspect-srpm')
        cls.prepare_results_dir()

        result = {'path': cls.get_checker_output_dir_short(), 'files': [], 'checks': {}}
        old_pkg = results_store.get_old_build()['srpm']
        new_pkg = results_store.get_new_build()['srpm']
        name = RpmHelper.split_nevra(os.path.basename(old_pkg))['name']
        outfile, pkg_data = cls.run_rpminspect(cls.results_dir, old_pkg, new_pkg)
        result['files'].append(os.path.basename(outfile))
        result['checks'][name] = pkg_data

        return result



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
from typing import Dict, List, Optional, cast
from xml.etree import ElementTree

from rebasehelper.exceptions import RebaseHelperError, CheckerNotFoundError
from rebasehelper.logger import CustomLogger
from rebasehelper.results_store import results_store
from rebasehelper.plugins.checkers import BaseChecker, CheckerCategory
from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.helpers.rpm_helper import RpmHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class PkgDiff(BaseChecker):
    """ Pkgdiff compare tool. """

    DEFAULT: bool = True
    CATEGORY: Optional[CheckerCategory] = CheckerCategory.RPM

    CMD: str = 'pkgdiff'
    CHECKER_TAGS: List[str] = ['added', 'removed', 'changed', 'moved', 'renamed']
    pkgdiff_results_filename: str = 'report'
    files_xml: str = 'files.xml'
    results_dir: str = ''
    results_dict: Dict[str, List[str]] = {}

    @classmethod
    def is_available(cls):
        try:
            return ProcessHelper.run_subprocess([cls.CMD, '--help'], output_file=ProcessHelper.DEV_NULL) == 0
        except (IOError, OSError):
            return False

    @classmethod
    def _get_rpm_info(cls, name, packages):
        if packages is None:
            return None
        basic_package = sorted(packages)[0]
        return getattr(RpmHelper.get_header_from_rpm(basic_package), name)

    @classmethod
    def _create_xml(cls, name, input_structure):
        """
        Function creates a XML format for pkgdiff command
        :param name: package name
        :param input_structure: structure provided by OutputLogger.get_build('new' or 'old')
        :return:
        """
        file_name = os.path.join(cls.results_dir, name + ".xml")
        if input_structure.get('version', '') == '':
            input_structure['version'] = cls._get_rpm_info('version', input_structure['rpm'])

        if input_structure.get('name', '') == '':
            input_structure['name'] = cls._get_rpm_info('name', input_structure['rpm'])

        tags = {'version': input_structure.get('version', ""),
                'group': input_structure.get('name', ''),
                'packages': input_structure.get('rpm', [])}
        lines = []
        for key, value in tags.items():
            new_value = value if isinstance(value, str) else '\n'.join(value)
            lines.append('<{0}>\n{1}\n</{0}>\n'.format(key, new_value))

        try:
            with open(file_name, 'w') as f:
                f.writelines(lines)
        except IOError as e:
            raise RebaseHelperError("Unable to create XML file for pkgdiff tool '{}'".format(file_name)) from e

        return file_name

    @classmethod
    def _remove_not_changed_files(cls):
        """
        Function removes all rows which were not changed
        """
        for tag in cls.CHECKER_TAGS:
            cls.results_dict[tag] = [x for x in cls.results_dict[tag] if not x.endswith('(0%)')]

    @classmethod
    def fill_dictionary(cls, result_dir, old_version=None, new_version=None):
        """
        Parsed files.xml and symbols.xml and fill dictionary
        :param result_dir: where should be stored file for pkgdiff
        :param old_version: old version of package
        :param new_version: new version of package
        :return:
        """
        XML_FILES = ['files.xml', 'symbols.xml']
        if old_version is None:
            old_version = results_store.get_old_build().get('version')
            if old_version == '':
                old_version = cls._get_rpm_info('version', results_store.get_old_build()['rpm'])
        if new_version is None:
            new_version = results_store.get_new_build().get('version')
            if new_version == '':
                new_version = cls._get_rpm_info('version', results_store.get_new_build()['rpm'])

        for tag in cls.CHECKER_TAGS:
            cls.results_dict[tag] = []
        for file_name in [os.path.join(result_dir, x) for x in XML_FILES]:
            logger.verbose('Processing %s file.', file_name)
            try:
                with open(file_name, "r") as f:
                    lines = ['<pkgdiff>']
                    lines.extend(f.readlines())
                    lines.append('</pkgdiff>')
                    pkgdiff_tree = ElementTree.fromstringlist(lines)
                    for tag in cls.CHECKER_TAGS:
                        for pkgdiff in pkgdiff_tree.findall('.//' + tag):
                            files = [x.strip() for x in pkgdiff.text.strip().split('\n')]
                            files = [x.replace(old_version, '*') for x in files]
                            files = [x.replace(new_version, '*') for x in files]
                            cls.results_dict[tag].extend(files)
            except IOError:
                continue

    @classmethod
    def _update_changed_moved(cls, key):
        updated_list = []
        for item in cls.results_dict[key]:
            fields = item.split(';')
            found = [x for x in cls.results_dict['changed'] if os.path.basename(fields[0]) in x]
            if not found:
                updated_list.append(item)
        return updated_list

    @classmethod
    def _remove_not_checked_files(cls, results_dict):
        """
        Function removes things which we don't care like
        ['.build-id', '.dwz']
        :return:
        """
        update_list = []
        removed_things = ['.build-id', '.dwz']
        for item in results_dict:
            removed = [x for x in removed_things if x in item]
            if removed:
                continue
            if item.endswith('.debug'):
                continue
            update_list.append(item)
        return update_list

    @classmethod
    def process_xml_results(cls, result_dir, old_version=None, new_version=None):
        """
        Function for filling dictionary with keys like 'added', 'removed'

        :return: dict = {'added': [list_of_added],
                         'removed': [list of removed],
                         'changed': [list of changed],
                         'moved': [list of moved]
                        }
        """
        cls.fill_dictionary(result_dir, old_version=old_version, new_version=new_version)

        # Remove all files which were not changed
        cls._remove_not_changed_files()
        for tag in cls.CHECKER_TAGS:
            cls.results_dict[tag] = cls._remove_not_checked_files(cls.results_dict[tag])

        added = [x for x in cls.results_dict['added'] if x not in cls.results_dict['removed']]
        removed = [x for x in cls.results_dict['removed'] if x not in cls.results_dict['added']]
        cls.results_dict['added'] = added
        cls.results_dict['removed'] = removed

        # remove unchanged files and remove things which are not checked
        # remove files from 'moved' if they are in 'changed' section
        cls.results_dict['moved'] = cls._update_changed_moved('moved')

        # Remove empty items
        return dict((k, v) for k, v in cls.results_dict.items() if v)

    @classmethod
    def run_check(cls, results_dir, **kwargs):
        """
        Compares old and new RPMs using pkgdiff
        :param results_dir result dir where are stored results
        """
        cls.results_dir = os.path.join(results_dir, cls.name)
        cls.prepare_results_dir()
        cls.pkgdiff_results_full_path_html = os.path.join(cls.results_dir, cls.pkgdiff_results_filename + '.html')

        cmd = [cls.CMD]
        cmd.append('-hide-unchanged')
        for version in ['old', 'new']:
            old = results_store.get_build(version)
            if old:
                file_name = cls._create_xml(version, input_structure=old)
                cmd.append(file_name)
        cmd.append('-extra-info')
        cmd.append(cls.results_dir)
        cmd.append('-report-path')
        cmd.append(cls.pkgdiff_results_full_path_html)
        try:
            ret_code = ProcessHelper.run_subprocess(cmd, output_file=ProcessHelper.DEV_NULL)
        except OSError as e:
            raise CheckerNotFoundError("Checker '{}' was not found or installed.".format(cls.name)) from e

        # From pkgdiff source code:
        # ret_code 0 means unchanged
        # ret_code 1 means Changed
        # other return codes means error
        if int(ret_code) != 0 and int(ret_code) != 1:
            raise RebaseHelperError('Execution of {} failed.\nCommand line is: {}'.format(cls.CMD, cmd))
        results_dict = cls.process_xml_results(cls.results_dir)
        lines: List[str] = []

        for key, val in results_dict.items():
            if val:
                if lines:
                    lines.append('')
                lines.append('Following files were {}:'.format(key))
                lines.extend(val)

        pkgdiff_report = os.path.join(cls.results_dir, cls.pkgdiff_results_filename + '.txt')
        try:
            with open(pkgdiff_report, "w") as f:
                f.write('\n'.join(lines))
        except IOError as e:
            raise RebaseHelperError("Unable to write result from {} to '{}'".format(cls.name, pkgdiff_report)) from e

        return dict(path=cls.get_checker_output_dir_short())

    @classmethod
    def format(cls, data):
        """
        Formats pkgdiff data to string
        :param data: pkgdiff data dictionary
        :return: formated pkgdiff list of strings
        """
        output_lines = [cls.get_underlined_title("pkgdiff")]
        output_lines.append("Details in {}:".format(data['path']))
        output_lines.append(" - {}.html".format(cls.pkgdiff_results_filename))
        output_lines.append(" - {}.txt".format(cls.pkgdiff_results_filename))
        return output_lines



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
from typing import cast
from distutils.util import strtobool

from rebasehelper.logger import CustomLogger


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class InputHelper:

    """Class for command line interaction with the user."""

    @staticmethod
    def get_message(message, default_yes=True, any_input=False):
        """Prompts a user with yes/no message and gets the response.

        Args:
            message (str): Prompt string.
            default_yes (bool): If the default value should be YES.
            any_input (bool): Whether to return default value regardless of input.

        Returns:
            bool: True or False, based on user's input.

        """
        if default_yes:
            choice = '[Y/n]'
        else:
            choice = '[y/N]'

        if any_input:
            msg = '{0} '.format(message)
        else:
            msg = '{0} {1}? '.format(message, choice)

        while True:
            user_input = input(msg).lower()

            if not user_input or any_input:
                return True if default_yes else False

            try:
                user_input = strtobool(user_input)
            except ValueError:
                logger.error('You have to type y(es) or n(o).')
                continue

            if any_input:
                return True
            else:
                return bool(user_input)



import cgi
import hashlib
import http.server
import io
import os
import posixpath
import ssl
import threading
import time
import urllib.parse

import pyftpdlib.authorizers
import pyftpdlib.handlers
import pyftpdlib.servers


class FTPServer:

    def __init__(self, port, root, report_size):
        class FTPHandlerNoSIZE(pyftpdlib.handlers.FTPHandler):
            proto_cmds = {k: v for k, v in pyftpdlib.handlers.proto_cmds.items() if k != 'SIZE'}

        authorizer = pyftpdlib.authorizers.DummyAuthorizer()
        authorizer.add_anonymous(root)
        handler = pyftpdlib.handlers.FTPHandler if report_size else FTPHandlerNoSIZE
        handler.authorizer = authorizer
        self.server = pyftpdlib.servers.FTPServer(('', port), handler)

    def serve(self):
        self.server.serve_forever()


class HTTPServer:

    def __init__(self, port, cert, root, report_size):
        class RequestHandler(http.server.BaseHTTPRequestHandler):

            def do_GET(self):
                path = self.path.split('?', 1)[0].split('#', 1)[0]
                path = urllib.parse.unquote(path)
                path = posixpath.normpath(path)
                path = os.path.join(root, path.lstrip('/'))
                try:
                    with open(path, 'rb') as f:
                        data = f.read()
                    self.send_response(200)
                    content_type = 'application/json' if 'versioneers' in path else 'application/octet-stream'
                    self.send_header('Content-Type', content_type)
                    self.send_header('Content-Transfer-Encoding', 'binary')
                    if report_size:
                        self.send_header('Content-Length', len(data))
                    self.end_headers()
                    self.wfile.write(data)
                except FileNotFoundError:
                    self.send_error(404)

            def do_POST(self):
                def dechunk(f):
                    bio = io.BytesIO()
                    while True:
                        chunksize = bytearray()
                        while not chunksize.endswith(b'\r\n'):
                            chunksize += f.read(1)
                        chunksize = chunksize.decode().split(':')[0]
                        chunksize = int(chunksize, 16)
                        if chunksize == 0:
                            break
                        chunk = f.read(chunksize)
                        assert(f.read(2) == b'\r\n')
                        bio.write(chunk)
                    bio.seek(0)
                    return bio

                def verify_hash(f, hashtype, hsh):
                    try:
                        chksum = hashlib.new(hashtype)
                    except ValueError:
                        return False
                    chksum.update(f.read())
                    return chksum.hexdigest() == hsh

                if self.headers.get('Transfer-Encoding') == 'chunked':
                    fp = dechunk(self.rfile)
                else:
                    fp = self.rfile
                data = cgi.FieldStorage(fp=fp, headers=self.headers,
                                        environ={'REQUEST_METHOD': 'POST'},
                                        # accept maximum of 10MB of data
                                        limit=10 * 1024 * 1024)
                try:
                    if 'filename' in data:
                        resp = b'Missing'
                        self.send_response(200)
                        self.send_header('Content-Type', 'text/plain')
                        self.send_header('Content-Length', len(resp))
                        self.end_headers()
                        self.wfile.write(resp)
                    else:
                        hashtype = [k for k in data.keys() if k.endswith('sum')][0]
                        hsh = data[hashtype].value
                        hashtype = hashtype.split('sum')[0]
                        if verify_hash(data['file'].file, hashtype, hsh):
                            self.send_response(204)
                            self.end_headers()
                        else:
                            self.send_error(500)
                except (KeyError, IndexError):
                    self.send_error(400)

        self.server = http.server.HTTPServer(('', port), RequestHandler)
        if cert:
            self.server.socket = ssl.wrap_socket(self.server.socket, certfile=cert, server_side=True)

    def serve(self):
        self.server.serve_forever()


def main():
    servers = [
        FTPServer(2100, '/srv', True),
        FTPServer(2101, '/srv', False),
        HTTPServer(8000, None, '/srv', True),
        HTTPServer(8001, None, '/srv', False),
        HTTPServer(4430, '/cert.pem', '/srv', True),
        HTTPServer(4431, '/cert.pem', '/srv', False),
    ]
    threads = [threading.Thread(target=s.serve) for s in servers[1:]]
    for t in threads:
        t.setDaemon(True)
        t.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
from typing import cast

from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.helpers.path_helper import PathHelper
from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.build_tools import MockTemporaryEnvironment, check_mock_privileges, get_mock_logfile_path
from rebasehelper.plugins.build_tools.rpm import BuildToolBase
from rebasehelper.exceptions import BinaryPackageBuildError


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class Mock(BuildToolBase):  # pylint: disable=abstract-method
    """
    Class representing Mock build tool.
    """

    DEFAULT: bool = True
    ACCEPTS_OPTIONS: bool = True

    CMD: str = 'mock'

    @classmethod
    def _build_rpm(cls, srpm, results_dir, rpm_results_dir, root=None, arch=None, builder_options=None):
        """Builds RPMs using mock.

        Args:
            srpm: Path to SRPM.
            results_dir: Path to directory where logs will be placed.
            rpm_results_dir: Path to directory where rpms will be placed.
            root: Path to where chroot will be built
            arch: Target architectures for the build.
            builder_options: Additional options for mock.

        Returns:
            Tuple where first element is list of paths to built RPMs and
            the second element is list of paths to logs.

        """
        logger.info("Building RPMs")
        output = os.path.join(results_dir, "mock_output.log")

        cmd = [cls.CMD, '--old-chroot', '--rebuild', srpm, '--resultdir', results_dir]
        if root is not None:
            cmd.extend(['--root', root])
        if arch is not None:
            cmd.extend(['--arch', arch])
        if builder_options is not None:
            cmd.extend(builder_options)

        if not check_mock_privileges():
            cmd = ['pkexec'] + cmd

        ret = ProcessHelper.run_subprocess(cmd, output_file=output)
        logs = []
        for log in PathHelper.find_all_files(results_dir, '*.log'):
            logs.append(os.path.join(rpm_results_dir, os.path.basename(log)))

        if ret == 0:
            return [f for f in PathHelper.find_all_files(results_dir, '*.rpm') if not f.endswith('.src.rpm')], logs
        else:
            logfile = get_mock_logfile_path(ret, rpm_results_dir, tmp_path=results_dir)
        raise BinaryPackageBuildError("Building RPMs failed!", rpm_results_dir, logfile=logfile, logs=logs)

    @classmethod
    def build(cls, spec, results_dir, srpm, **kwargs):
        """
        Builds the RPMs using mock

        :param spec: SpecFile object
        :param results_dir: absolute path to directory where results will be stored
        :param srpm: absolute path to SRPM
        :param root: mock root used for building
        :param arch: architecture to build the RPM for
        :return: dict with:
                 'rpm' -> list with absolute paths to RPMs
                 'logs' -> list with absolute paths to logs
        """
        sources = spec.get_sources()
        patches = [p.path for p in spec.get_patches()]
        with MockTemporaryEnvironment(sources, patches, spec.path, results_dir) as tmp_env:
            env = tmp_env.env()
            tmp_results_dir = env.get(MockTemporaryEnvironment.TEMPDIR_RESULTS)
            rpms, logs = cls._build_rpm(srpm, tmp_results_dir, results_dir,
                                        builder_options=cls.get_builder_options(**kwargs))
            # remove SRPM - side product of building RPM
            tmp_srpm = PathHelper.find_first_file(tmp_results_dir, "*.src.rpm")
            if tmp_srpm is not None:
                os.unlink(tmp_srpm)

        logger.info("Building RPMs finished successfully")

        rpms = [os.path.join(results_dir, os.path.basename(f)) for f in rpms]
        logger.verbose("Successfully built RPMs: '%s'", str(rpms))
        logger.verbose("logs: '%s'", str(logs))

        return dict(rpm=rpms, logs=logs)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import argparse
import logging
import os
import re
import sys
from typing import cast, Dict, List

from rebasehelper import VERSION
from rebasehelper.options import OPTIONS, traverse_options
from rebasehelper.constants import PROGRAM_DESCRIPTION, NEW_ISSUE_LINK, LOGS_DIR, TRACEBACK_LOG, DEBUG_LOG
from rebasehelper.application import Application
from rebasehelper.logger import CustomLogger, LoggerHelper
from rebasehelper.exceptions import RebaseHelperError
from rebasehelper.helpers.console_helper import ConsoleHelper
from rebasehelper.helpers.bugzilla_helper import BugzillaHelper
from rebasehelper.config import Config
from rebasehelper.argument_parser import CustomArgumentParser, CustomHelpFormatter, CustomAction
from rebasehelper.plugins.plugin_manager import plugin_manager


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class CLI:
    """ Class for processing data from commandline """

    @staticmethod
    def build_parser(available_choices_only=False):
        parser = CustomArgumentParser(description=PROGRAM_DESCRIPTION,
                                      formatter_class=CustomHelpFormatter)

        group = None
        current_group = 0
        for option in traverse_options(OPTIONS + plugin_manager.get_options()):
            available_choices = option.pop("available_choices", option.get("choices"))
            if available_choices_only:
                option["choices"] = available_choices

            option_kwargs = dict(option)
            for key in ("group", "default", "name"):
                if key in option_kwargs:
                    del option_kwargs[key]

            if "group" in option:
                if not group or current_group != option["group"]:
                    current_group = option["group"]
                    group = parser.add_mutually_exclusive_group()
                actions_container = group
            else:
                actions_container = parser

            # default is set to SUPPRESS to prevent arguments which were not specified on command line from being
            # added to namespace. This allows rebase-helper to determine which arguments are used with their
            # default value. This is later used for merging CLI arguments with config.
            actions_container.add_argument(*option["name"], action=CustomAction, default=argparse.SUPPRESS,
                                           actual_default=option.get("default"), **option_kwargs)

        return parser

    def __init__(self, args=None):
        """parse arguments"""
        if args is None:
            args = sys.argv[1:]
        # sanitize builder options to prevent ArgumentParser from processing them
        for opt in ['--builder-options', '--srpm-builder-options']:
            try:
                i = args.index(opt)
                args[i:i+2] = ['='.join(args[i:i+2])]
            except ValueError:
                continue
        self.parser = CLI.build_parser(available_choices_only=True)
        self.args = self.parser.parse_args(args)

    def __getattr__(self, name):
        try:
            return getattr(self.args, name)
        except AttributeError:
            return object.__getattribute__(self, name)


class CliHelper:

    @classmethod
    def convert_macros_to_dict(cls, macros_list: List[str]) -> Dict[str, str]:
        """Converts macros from CLI to a dictionary.

        Args:
            macros_list: List of macros in the format 'MACRO EXPRESSION'.

        Returns:
            The converted macros, MACRO are keys, EXPRESSION are values.

        Raises:
            RebaseHelperError if the macros don't follow the correct format.

        """
        macros = {}
        for macro in macros_list:
            match = re.match(r"\s*%?(?P<macro>\S+)\s+(?P<expr>\S.*)\s*", macro)
            if not match:
                raise RebaseHelperError('Wrong macro format used ({}). Use \'MACRO EXPR\' instead'.format(macro))

            macros[match.group('macro')] = match.group('expr')

        return macros

    @classmethod
    def run(cls):
        results_dir = None
        start_dir = os.getcwd()
        try:
            LoggerHelper.setup_memory_handler()
            main_handler, output_tool_handler = LoggerHelper.create_stream_handlers()
            cli = CLI()
            if hasattr(cli, 'version'):
                print(VERSION)
                sys.exit(0)

            logger.info('rebase-helper version %s', VERSION)

            config = Config(getattr(cli, 'config-file', None))
            config.merge(cli)
            for handler in [main_handler, output_tool_handler]:
                handler.set_terminal_background(config.background)
            if config.verbose == 0:
                main_handler.setLevel(logging.INFO)
            elif config.verbose == 1:
                main_handler.setLevel(CustomLogger.VERBOSE)
            else:
                main_handler.setLevel(logging.DEBUG)
            ConsoleHelper.use_colors = ConsoleHelper.should_use_colors(config)
            if config.bugzilla_id:
                repo_path, config.config['sources'] = BugzillaHelper.prepare_rebase_repository(config.bugzilla_id)
                try:
                    os.chdir(repo_path)
                except OSError as e:
                    raise RebaseHelperError('Could not change directory to the cloned repository') from e
                # update relative paths in config
                for option in ('results_dir', 'workspace_dir'):
                    path = getattr(config, option)
                    if path and not os.path.isabs(path):
                        config.config[option] = os.path.join(os.path.relpath(start_dir, os.getcwd()), path)

            if config.config['rpmmacros']:
                macros = ' -D ' + ' -D '.join('"{}"'.format(s) for s in config.config['rpmmacros'])
                for option in ('builder_options', 'srpm_builder_options'):
                    if config.config[option]:
                        config.config[option] += macros
                    else:
                        config.config[option] = macros

            config.config['rpmmacros'] = cls.convert_macros_to_dict(config.rpmmacros)
            execution_dir, results_dir = Application.setup(config)
            app = Application(config, start_dir, execution_dir, results_dir)
            app.run()
        except KeyboardInterrupt:
            logger.info('Interrupted by user')
        except RebaseHelperError as e:
            if e.msg:
                logger.error('%s', e.msg)
            else:
                logger.error('%s', str(e))
            sys.exit(1)
        except SystemExit as e:
            sys.exit(e.code)
        except BaseException:
            logger_traceback: CustomLogger = cast(CustomLogger, logging.getLogger('rebasehelper.traceback'))
            logger_traceback.propagate = False
            logger_traceback.setLevel(CustomLogger.TRACE)
            if results_dir:
                debug_log = os.path.join(results_dir, LOGS_DIR, DEBUG_LOG)
                traceback_log = os.path.join(results_dir, LOGS_DIR, TRACEBACK_LOG)
                logger.error('rebase-helper failed due to an unexpected error. Please report this problem'
                             '\nusing the following link: %s'
                             '\nand include the content of'
                             '\n\'%s\' and'
                             '\n\'%s\''
                             '\nin the report.'
                             '\nThank you!',
                             NEW_ISSUE_LINK, debug_log, traceback_log)
                LoggerHelper.add_file_handler(logger_traceback, traceback_log)
            else:
                logger.error('rebase-helper failed due to an unexpected error. Please report this problem'
                             '\nusing the following link: %s'
                             '\nand include the following traceback in the report.'
                             '\nThank you!',
                             NEW_ISSUE_LINK)

                LoggerHelper.add_stream_handler(logger_traceback, CustomLogger.TRACE)
            logger_traceback.trace('', exc_info=1)
            sys.exit(1)

        sys.exit(0)


if __name__ == '__main__':
    x = CLI()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import pytest  # type: ignore


@pytest.mark.public_api
class TestSpecContent:
    def test_section(self, spec_object):
        assert isinstance(spec_object.spec_content.section('%install'), list)
        assert spec_object.spec_content.section(name='%nonexistent') is None

    def test_sections(self, spec_object):
        assert isinstance(spec_object.spec_content.sections, list)
        for section in spec_object.spec_content.sections:
            assert isinstance(section, tuple)
            assert isinstance(section[1], list)

    def test_replace_section(self, spec_object):
        section = '%install'
        replacement = ['test', '']
        assert spec_object.spec_content.replace_section(section, replacement) is True
        assert spec_object.spec_content.replace_section(name=section, content=replacement) is True
        assert spec_object.spec_content.replace_section(name='%notasection', content=[]) is False



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os
from typing import List
from textwrap import dedent

import pytest  # type: ignore
import rpm  # type: ignore

from rebasehelper.tags import Tag
from rebasehelper.specfile import SpecFile


@pytest.fixture(autouse=True)
def rpm_cleanup():
    # Some tests change macro values. This can influence other tests.
    # Prevent it by resetting the macros before each test.
    rpm.reloadConfig()


class TestSpecFile:
    NAME: str = 'test'
    VERSION: str = '1.0.2'
    OLD_ARCHIVE: str = NAME + '-' + VERSION + '.tar.xz'
    SOURCE_1: str = 'source-tests.sh'
    SOURCE_2: str = ''
    SOURCE_5: str = 'documentation.tar.xz'
    SOURCE_6: str = 'misc.zip'
    PATCH_1: str = 'test-testing.patch'
    PATCH_2: str = 'test-testing2.patch'
    PATCH_3: str = 'test-testing3.patch'
    PATCH_4: str = 'test-testing4.patch'

    TEST_FILES: List[str] = [
        OLD_ARCHIVE,
        SOURCE_1,
        SOURCE_5,
        SOURCE_6,
        PATCH_1,
        PATCH_2,
        PATCH_3,
        PATCH_4,
    ]

    def test_get_release(self, spec_object):
        assert spec_object.get_release() == '34'

    def test_set_release(self, spec_object):
        spec_object.set_release('0.1')
        assert spec_object.get_release() == '0.1'
        spec_object.set_release('22')
        assert spec_object.get_release() == '22'

    def test_set_version(self, spec_object):
        NEW_VERSION = '1.2.3.4.5'
        spec_object.set_version(NEW_VERSION)
        spec_object.save()
        assert spec_object.header.version == NEW_VERSION

    @pytest.mark.parametrize('spec_attributes, sources', [
        (
            {
                'spec_content': dedent("""\
                    Source:     ftp://ftp.test.org/%{name}-%{version}.tar.xz
                    Source1:    source-tests.sh
                    Source2:    ftp://test.com/test-source.sh
                    #Source3:    source-tests.sh
                    """)
            },
            [
                'ftp://ftp.test.org/%{name}-%{version}.tar.xz',
                'source-tests.sh',
                'ftp://test.com/test-source.sh',
                None
            ]
        )
    ])
    def test__get_raw_source_string(self, mocked_spec_object, sources):
        # pylint: disable=protected-access
        for i, source in enumerate(sources):
            assert mocked_spec_object._get_raw_source_string(i) == source

    def test_old_tarball(self, spec_object):
        assert spec_object.get_archive() == self.OLD_ARCHIVE

    def test_get_sources(self, workdir, spec_object):
        sources = [self.SOURCE_1, self.SOURCE_5, self.SOURCE_6, self.OLD_ARCHIVE]
        sources = [os.path.join(workdir, f) for f in sources]
        assert len(set(sources).intersection(set(spec_object.get_sources()))) == 4
        # The Source0 has to be always in the beginning
        assert spec_object.get_archive() == 'test-1.0.2.tar.xz'

    def test_get_patches(self, workdir, spec_object):
        expected_patches = {0: [os.path.join(workdir, self.PATCH_1), 1],
                            1: [os.path.join(workdir, self.PATCH_2), 2],
                            2: [os.path.join(workdir, self.PATCH_3), 3],
                            3: [os.path.join(workdir, self.PATCH_4), 4]}
        patches = {}
        for index, p in enumerate(spec_object.get_patches()):
            patches[index] = [p.path, p.index]
        assert patches == expected_patches

    def test_split_version_string(self):
        assert SpecFile.split_version_string('1.0.1', '1.0.1') == ('1.0.1', None)
        assert SpecFile.split_version_string('1.0.1b1', '1.0.1') == ('1.0.1', 'b1')
        assert SpecFile.split_version_string('1.0.1rc1', '1.0.1') == ('1.0.1', 'rc1')
        assert SpecFile.split_version_string('1.1.3-rc6', '1.1.3') == ('1.1.3', 'rc6')
        assert SpecFile.split_version_string('1.1.3_rc6', '1.1.3') == ('1.1.3', 'rc6')
        assert SpecFile.split_version_string('1.1.3~rc6', '1.1.3') == ('1.1.3', 'rc6')
        assert SpecFile.split_version_string('1.1.1d', '1.1.1c') == ('1.1.1d', None)

    def test_extract_version_from_archive_name(self):
        # Basic tests
        assert SpecFile.extract_version_from_archive_name('test-1.0.1.tar.gz', '') == '1.0.1'
        assert SpecFile.extract_version_from_archive_name('/home/user/test-1.0.1.tar.gz', '') == '1.0.1'
        assert SpecFile.extract_version_from_archive_name('test-1.0.1.tar.gz',
                                                          'ftp://ftp.test.org/test-%{version}.tar.gz') == '1.0.1'
        assert SpecFile.extract_version_from_archive_name('/home/user/test-1.0.1.tar.gz',
                                                          'ftp://ftp.test.org/test-%{version}.tar.gz') == '1.0.1'
        # Real world tests
        name = 'http://www.cups.org/software/%{version}/cups-%{version}-source.tar.bz2'
        assert SpecFile.extract_version_from_archive_name('cups-1.7.5-source.tar.bz2',
                                                          name) == '1.7.5'
        name = 'ftp://ftp.isc.org/isc/bind9/%{VERSION}/bind-%{VERSION}.tar.gz'
        assert SpecFile.extract_version_from_archive_name('bind-9.9.5rc2.tar.gz',
                                                          name) == '9.9.5rc2'
        name = 'http://www.thekelleys.org.uk/dnsmasq/%{?extrapath}%{name}-%{version}%{?extraversion}.tar.xz'
        assert SpecFile.extract_version_from_archive_name('dnsmasq-2.69rc1.tar.xz',
                                                          name) == '2.69rc1'
        name = 'http://downloads.sourceforge.net/%{name}/%{name}-%{version}%{?prever:-%{prever}}.tar.xz'
        assert SpecFile.extract_version_from_archive_name('log4cplus-1.1.3-rc3.tar.xz',
                                                          name) == '1.1.3-rc3'
        name = 'http://downloads.sourceforge.net/%{name}/%{name}-%{version}%{?prever:_%{prever}}.tar.xz'
        assert SpecFile.extract_version_from_archive_name('log4cplus-1.1.3_rc3.tar.xz',
                                                          name) == '1.1.3_rc3'
        name = 'http://download.gnome.org/sources/libsigc++/%{release_version}/libsigc++-%{version}.tar.xz'
        assert SpecFile.extract_version_from_archive_name('libsigc++-2.10.0.tar.xz',
                                                          name) == '2.10.0'

    @pytest.mark.parametrize('spec_attributes, main_files', [
        (
            {
                'spec_content': dedent("""\
                    %files
                    %files -n test
                    %files test
                    """)
            },
            '%files'
        ),
        (
            {
                'spec_content': dedent("""\
                    %files -n test
                    """)
            },
            None
        ),
    ], ids=[
        'has-main-files',
        'no-main-files',
    ])
    def test_get_main_files_section(self, mocked_spec_object, main_files):
        assert mocked_spec_object.get_main_files_section() == main_files

    @pytest.mark.parametrize('spec_attributes, is_enabled', [
        (
            {
                'spec_content': dedent("""\
                    %check
                    make test
                    """)
            },
            True
        ),
        (
            {
                'spec_content': dedent("""\
                    %check
                    # disabled test
                    # make test
                    """)
            },
            False
        ),
    ], ids=[
        'is_enabled',
        'is_disabled',
    ])
    def test_is_test_suite_enabled(self, mocked_spec_object, is_enabled):
        assert mocked_spec_object.is_test_suite_enabled() is is_enabled

    def test_set_extra_version(self, spec_object):
        spec_object.set_version('1.0.3')
        spec_object.set_extra_version('beta1', True)
        assert spec_object.get_release() == '0.1.beta1'
        spec_object.set_extra_version('beta2', False)
        assert spec_object.get_release() == '0.2.beta2'
        spec_object.set_extra_version('rc', False)
        assert spec_object.get_release() == '0.3.rc'
        spec_object.set_version('1.0.4')
        spec_object.set_extra_version(None, True)
        assert spec_object.get_release() == '1'
        spec_object.set_extra_version('g1234567', False)
        assert spec_object.get_release() == '2.g1234567'

    @pytest.mark.parametrize('spec_attributes', [
        {
            'spec_content': dedent("""\
                %global release 34
                %global release_str %{release}%{?dist}

                Name:    test
                Version: 1.0.2
                Release: %{release_str}

                %prep
                %setup -q -c -a 5
                """),
            'get_release': lambda: '34',
            'macros':
                {
                    'name': {'value': 'test', 'level': -3},
                    'version': {'value': '1.0.2', 'level': -3},
                },
        }
    ])
    def test_update_setup_dirname(self, mocked_spec_object):
        mocked_spec_object.set_extra_version('rc1', False)

        prep = mocked_spec_object.spec_content.section('%prep')
        mocked_spec_object.update_setup_dirname('test-1.0.2')
        assert mocked_spec_object.spec_content.section('%prep') == prep

        mocked_spec_object.update_setup_dirname('test-1.0.2rc1')
        prep = mocked_spec_object.spec_content.section('%prep')
        setup = [l for l in prep if l.startswith('%setup')][0]
        assert '-n %{name}-%{version}rc1' in setup

        mocked_spec_object.update_setup_dirname('test-1.0.2-rc1')
        prep = mocked_spec_object.spec_content.section('%prep')
        setup = [l for l in prep if l.startswith('%setup')][0]
        assert '-n %{name}-%{version}-rc1' in setup

    def test_find_archive_target_in_prep(self, spec_object):
        target = spec_object.find_archive_target_in_prep('documentation.tar.xz')
        assert target == 'test-1.0.2'
        target = spec_object.find_archive_target_in_prep('misc.zip')
        assert target == 'test-1.0.2/misc'

    @pytest.mark.parametrize('spec_attributes, kwargs, expected_content', [
        (
            {
                'removed_patches': [],
                'spec_content': dedent("""\
                    Patch0:    0.patch
                    Patch1:    1.patch

                    %patchlist
                    2.patch
                    3.patch

                    %prep
                    %patch0 -p0
                    %patch1 -p1
                    %patch2 -p2
                    %patch3 -p3
                    """),
            },
            {
                'patches':
                    {
                        'deleted': ['2.patch'],
                        'inapplicable': ['1.patch'],
                    },
                'disable_inapplicable': False,
            },
            dedent("""\
                Patch0:    0.patch
                Patch1:    1.patch

                %patchlist
                3.patch

                %prep
                %patch0 -p0
                # The following patch contains conflicts
                %patch1 -p1
                %patch2 -p3
                """)
        ),
        (
            {
                'removed_patches': [],
                'spec_content': dedent("""\
                    Patch0:    0.patch
                    Patch1:    1.patch

                    %patchlist
                    2.patch
                    3.patch

                    %prep
                    %patch0 -p0
                    %patch1 -p1
                    %patch2 -p2
                    %patch3 -p3
                    """),
            },
            {
                'patches':
                    {
                        'deleted': ['2.patch'],
                        'inapplicable': ['1.patch'],
                    },
                'disable_inapplicable': True,
            },
            dedent("""\
        Patch0:    0.patch
        #Patch1:    1.patch

        %patchlist
        3.patch

        %prep
        %patch0 -p0
        # The following patch contains conflicts
        #%%patch1 -p1
        %patch1 -p3
        """),
        ),
        (
            {
                'removed_patches': [],
                'spec_content': dedent("""\
                    Patch0:     0.patch
                    
                    
                    # Patch comment
                    # line2
                    Patch1:     1.patch
                    
                    Patch2:     2.patch
                    """),
            },
            {
                'patches':
                    {
                        'deleted': ['1.patch'],
                    },
                'disable_inapplicable': False,
            },
            dedent("""\
            Patch0:     0.patch
            
            Patch2:     2.patch
            """),
        ),
    ], ids=[
        'do_not_disable_inapplicable',
        'disable_inapplicable',
        'comments_and_blank_lines',
    ])
    def test_write_updated_patches(self, mocked_spec_object, kwargs, expected_content):
        mocked_spec_object.write_updated_patches(**kwargs)
        assert expected_content == str(mocked_spec_object.spec_content)

    @pytest.mark.parametrize('spec_attributes', [
        {
            'spec_content': 'Patch5: rebase-helper-results/rebased-sources/test-testing5.patch\n'
        }
    ])
    def test_update_paths_to_sources_and_patches(self, mocked_spec_object):
        line = [l for l in mocked_spec_object.spec_content.section('%package') if l.startswith('Patch5')][0]
        assert 'rebased-sources' in line

        mocked_spec_object.update_paths_to_sources_and_patches()

        line = [l for l in mocked_spec_object.spec_content.section('%package') if l.startswith('Patch5')][0]
        assert 'rebased-sources' not in line

    def test_tags(self, spec_object):
        # sanity check
        assert spec_object.tag('Name') == Tag(0, '%package', 16, 'Name', (6, 10), True)
        # no workaround
        assert spec_object.tag('Patch100') is None
        assert spec_object.tag('Patch101') is not None
        assert spec_object.tag('Patch102') is None
        assert spec_object.get_raw_tag_value('Patch101') == 'no_workaround.patch'
        # workaround
        spec_object.predefined_macros = {'use_workaround': '1'}
        spec_object.update()
        assert spec_object.tag('Patch100') is not None
        assert spec_object.tag('Patch101') is not None
        assert spec_object.tag('Patch102') is not None
        assert spec_object.get_raw_tag_value('Patch100') == 'workaround_base.patch'
        assert spec_object.get_raw_tag_value('Patch101') == 'workaround_1.patch'
        assert spec_object.get_raw_tag_value('Patch102') == 'workaround_2.patch'

    @pytest.mark.parametrize('preserve_macros', [
        False,
        True,
    ], ids=[
        'ignoring_macros',
        'preserving_macros',
    ])
    @pytest.mark.parametrize('tag, value, lines, lines_preserve', [
        (
            'Summary',
            'A testing SPEC file',
            [
                '%{!?specfile: %global specfile spec file}',
                '%global summary %{?longsum}%{!?longsum:A testing %{specfile}}',
                'Summary: A testing SPEC file',
            ],
            [
                '%{!?specfile: %global specfile SPEC file}',
                '%global summary %{?longsum}%{!?longsum:A testing %{specfile}}',
                'Summary: %{summary}',
            ],
        ),
        (
            'Version',
            '1.1.8',
            [
                '%global version_major 1',
                '%global version_minor 0',
                '%global version_patch 2',
                '%global version_major_minor %{version_major}.%{version_minor}',
                '%global version %{version_major_minor}.%{version_patch}',
                'Version: 1.1.8',
            ],
            [
                '%global version_major 1',
                '%global version_minor 1',
                '%global version_patch 8',
                '%global version_major_minor %{version_major}.%{version_minor}',
                '%global version %{version_major_minor}.%{version_patch}',
                'Version: %{version}',
            ],
        ),
        (
            'Release',
            '42%{?dist}',
            [
                '%global release 34',
                '%global release_str %{release}%{?dist}',
                'Release: 42%{?dist}',
            ],
            [
                '%global release 42',
                '%global release_str %{release}%{?dist}',
                'Release: %{release_str}',
            ],
        ),
        (
            'Source8',
            'https://github.com/rebase-helper/rebase-helper/archive/'
            'b0ed0b235bd5ea295fc897e1e2e8e6b6637f2c2d/'
            'rebase-helper-b0ed0b235bd5ea295fc897e1e2e8e6b6637f2c2d.tar.gz',
            [
                '%global project rebase-helper',
                '%global commit d70cb5a2f523db5b6088427563531f43b7703859',
                'Source8: https://github.com/rebase-helper/rebase-helper/archive/'
                'b0ed0b235bd5ea295fc897e1e2e8e6b6637f2c2d/'
                'rebase-helper-b0ed0b235bd5ea295fc897e1e2e8e6b6637f2c2d.tar.gz',
            ],
            [
                '%global project rebase-helper',
                '%global commit b0ed0b235bd5ea295fc897e1e2e8e6b6637f2c2d',
                'Source8: https://github.com/%{project}/%{project}/archive/%{commit}/%{project}-%{commit}.tar.gz',
            ],
        ),
        (
                'Patch1000',
                '0.8.b5%{?dist}',
                [
                    '%global prever b4',
                    'Patch1000: 0.8.b5%{?dist}',
                ],
                [
                    '%global prever b5',
                    'Patch1000: 0.8.%{?prever}%{?dist}',
                ],
        ),
        (
                'Patch1001',
                '1.22.2',
                [
                    '%global branch 1.22',
                    'Patch1001: 1.22.2',
                ],
                [
                    '%global branch 1.22',
                    'Patch1001: %{branch}.2',
                ],
        ),
    ], ids=[
        'Summary=>"A testing SPEC file..."',
        'Version=>"1.1.8"',
        'Release=>"42%{?dist}"',
        'Source8=>"https://github.com/rebase-helper/rebase-helper/archive/..."',
        'Patch1000=>"0.8.b5%{?dist}"',
        'Patch1001=>"1.22.2"',
    ])
    def test_set_tag(self, spec_object, preserve_macros, tag, value, lines, lines_preserve):
        spec_object.set_tag(tag, value, preserve_macros=preserve_macros)
        for line in lines_preserve if preserve_macros else lines:
            assert line in spec_object.spec_content.section('%package')



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import logging.handlers
import os
from typing import Dict, List, Optional, Tuple

from rebasehelper.helpers.console_helper import ConsoleHelper
from rebasehelper import constants


class CustomLogger(logging.Logger):

    TRACE: int = logging.DEBUG + 1
    VERBOSE: int = logging.DEBUG + 2
    SUCCESS: int = logging.INFO + 5
    HEADING: int = logging.INFO + 6
    IMPORTANT: int = logging.INFO + 7

    _nameToLevel: Dict[str, int] = {
        'TRACE': TRACE,
        'VERBOSE': VERBOSE,
        'SUCCESS': SUCCESS,
        'HEADING': HEADING,
        'IMPORTANT': IMPORTANT,
    }

    def __init__(self, name, level=logging.NOTSET):
        super(CustomLogger, self).__init__(name, level)

        for lev, severity in self._nameToLevel.items():
            logging.addLevelName(severity, lev)

    def __getattr__(self, level):
        severity = self._nameToLevel.get(level.upper())

        def log(message, *args, **kwargs):
            if self.isEnabledFor(severity):
                self._log(severity, message, args, **kwargs)

        if severity:
            return log

        raise AttributeError


class ColorizingStreamHandler(logging.StreamHandler):
    colors: Dict[str, Dict[int, Dict[str, Optional[str]]]] = {
        'dark': {
            logging.DEBUG: {'fg': 'brightblack', 'bg': 'default', 'style': None},
            CustomLogger.TRACE: {'fg': 'red', 'bg': 'default', 'style': None},
            CustomLogger.VERBOSE: {'fg': 'brightblack', 'bg': 'default', 'style': None},
            logging.INFO: {'fg': 'default', 'bg': 'default', 'style': None},
            CustomLogger.SUCCESS: {'fg': 'green', 'bg': 'default', 'style': None},
            CustomLogger.HEADING: {'fg': 'yellow', 'bg': 'default', 'style': None},
            CustomLogger.IMPORTANT: {'fg': 'red', 'bg': 'default', 'style': None},
            logging.WARNING: {'fg': 'yellow', 'bg': 'default', 'style': None},
            logging.ERROR: {'fg': 'red', 'bg': 'default', 'style': 'bold'},
            logging.CRITICAL: {'fg': 'white', 'bg': 'red', 'style': 'bold'},
        },
        'light': {
            logging.DEBUG: {'fg': 'brightblack', 'bg': 'default', 'style': None},
            CustomLogger.TRACE: {'fg': 'red', 'bg': 'default', 'style': None},
            CustomLogger.VERBOSE: {'fg': 'brightblack', 'bg': 'default', 'style': None},
            logging.INFO: {'fg': 'default', 'bg': 'default', 'style': None},
            CustomLogger.SUCCESS: {'fg': 'green', 'bg': 'default', 'style': None},
            CustomLogger.HEADING: {'fg': 'blue', 'bg': 'default', 'style': None},
            CustomLogger.IMPORTANT: {'fg': 'red', 'bg': 'default', 'style': None},
            logging.WARNING: {'fg': 'blue', 'bg': 'default', 'style': None},
            logging.ERROR: {'fg': 'red', 'bg': 'default', 'style': 'bold'},
            logging.CRITICAL: {'fg': 'white', 'bg': 'red', 'style': 'bold'},
        },
    }

    terminal_background: str = 'dark'

    def set_terminal_background(self, background):
        if background == 'auto':
            self.terminal_background = ConsoleHelper.detect_background()
        else:
            self.terminal_background = background

    def emit(self, record):
        try:
            message = self.format(record)
            level_settings = self.colors[self.terminal_background].get(record.levelno, {})
            ConsoleHelper.cprint(message, **level_settings)
            self.flush()
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)


class MemoryHandler(logging.handlers.BufferingHandler):
    """BufferingHandler with infinite capacity"""

    buffer: List[logging.LogRecord]  # until this is added to typeshed: https://github.com/python/typeshed/pull/3402

    def __init__(self) -> None:
        super(MemoryHandler, self).__init__(0)

    def shouldFlush(self, record: logging.LogRecord) -> bool:
        return False

    def replay_into(self, target: logging.Handler) -> None:
        self.acquire()
        try:
            for record in self.buffer:
                if record.levelno >= target.level:
                    target.handle(record)
        finally:
            self.release()


class LoggerHelper:
    """Helper class for setting up a logger."""

    memory_handler: Optional[MemoryHandler] = None

    @classmethod
    def setup_memory_handler(cls) -> None:
        if cls.memory_handler:
            # only one memory handler is allowed
            return
        cls.memory_handler = MemoryHandler()
        logger = logging.getLogger('rebasehelper')
        logger.addHandler(cls.memory_handler)

    @classmethod
    def remove_memory_handler(cls) -> None:
        if cls.memory_handler:
            logger = logging.getLogger('rebasehelper')
            logger.removeHandler(cls.memory_handler)
        cls.memory_handler = None

    @staticmethod
    def add_stream_handler(logger: logging.Logger, level: Optional[int] = None,
                           formatter: Optional[logging.Formatter] = None) -> ColorizingStreamHandler:
        """Adds stream handler to the given logger.

        Args:
            logger: Logger object to add the handler to.
            level: Severity threshold.
            formatter: Formatter object used to format logged messages.

        Returns:
            Created stream handler instance.

        """
        console_handler = ColorizingStreamHandler()
        if level:
            console_handler.setLevel(level)
        if formatter:
            console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
        return console_handler

    @staticmethod
    def add_file_handler(logger: logging.Logger, path: str, formatter: Optional[logging.Formatter] = None,
                         level: Optional[int] = None) -> Optional[logging.FileHandler]:
        """Adds file handler to the given logger.

        Args:
            logger: Logger object to add the handler to.
            path: Path to a log file.
            formatter: Formatter object used to format logged messages.
            level: Severity threshold.

        Returns:
            Created file handler instance or None if creation failed.

        """
        try:
            file_handler = logging.FileHandler(path, 'w')
            if level:
                file_handler.setLevel(level)
            if formatter:
                file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)
        except (IOError, OSError):
            logger.warning('Can not create log in %s', path)
            return None
        else:
            return file_handler

    @classmethod
    def create_file_handlers(cls, results_dir: str) -> List[logging.FileHandler]:
        """Creates rebase-helper file handlers.

        Args:
            results_dir: Path to rebase-helper-results directory.

        Returns:
            List of created file handler instances.

        """
        logs_dir = os.path.join(results_dir, constants.LOGS_DIR)
        # the logs directory can already exist
        os.makedirs(logs_dir, exist_ok=True)
        logger = logging.getLogger('rebasehelper')

        # first remove any existing file handlers
        logger.handlers = [h for h in logger.handlers if not isinstance(h, logging.FileHandler)]

        log_formatter = logging.Formatter('%(message)s')
        debug_log_formatter = logging.Formatter('%(asctime)s %(filename)s:%(lineno)s %(funcName)s: %(message)s')

        debug_log = os.path.join(logs_dir, constants.DEBUG_LOG)
        debug = cls.add_file_handler(logger, debug_log, debug_log_formatter, logging.DEBUG)
        verbose_log = os.path.join(logs_dir, constants.VERBOSE_LOG)
        verbose = cls.add_file_handler(logger, verbose_log, log_formatter, CustomLogger.VERBOSE)
        info_log = os.path.join(logs_dir, constants.INFO_LOG)
        info = cls.add_file_handler(logger, info_log, log_formatter, logging.INFO)

        if cls.memory_handler:
            # initialize the log files with what has been recorded in memory until now
            for handler in (debug, verbose, info):
                if handler:
                    cls.memory_handler.replay_into(handler)
            cls.remove_memory_handler()

        return [h for h in (debug, verbose, info) if h]

    @classmethod
    def create_stream_handlers(cls) -> Tuple[ColorizingStreamHandler, ColorizingStreamHandler]:
        logger = logging.getLogger('rebasehelper')
        formatter = logging.Formatter('%(levelname)s: %(message)s')
        main = cls.add_stream_handler(logger, logging.INFO, formatter)

        logger_summary = logging.getLogger('rebasehelper.summary')
        logger_summary.propagate = False
        summary = cls.add_stream_handler(logger_summary)

        logger_report = logging.getLogger('rebasehelper.report')
        logger_report.propagate = False
        return main, summary

    @classmethod
    def remove_file_handlers(cls, handlers: List[logging.FileHandler]) -> None:
        """Removes rebase-helper file handlers.

        Args:
            handlers: List of file handlers to remove.

        """
        logger = logging.getLogger('rebasehelper')
        for handler in handlers:
            logger.removeHandler(handler)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

from typing import Dict, Optional, Type

import pkg_resources

from rebasehelper.plugins.plugin import Plugin


class PluginLoader:
    @classmethod
    def load(cls, entrypoint, manager):
        result: Dict[str, Optional[Type[Plugin]]] = {}
        for ep in pkg_resources.iter_entry_points(entrypoint):
            result[ep.name] = None
            try:
                plugin = ep.load()
            except BaseException:
                # skip broken plugin
                continue
            try:
                if not issubclass(plugin, Plugin):
                    raise TypeError
            except TypeError:
                # skip broken plugin
                continue
            else:
                plugin.name = ep.name
                # Some plugins require access to other plugins. Avoid cyclic
                # imports by setting manager as an attribute.
                plugin.manager = manager
            result[ep.name] = plugin
        return result



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os
import shutil

import pytest  # type: ignore

from rebasehelper.specfile import SpecFile
from rebasehelper.tags import Tag
from rebasehelper.tests.conftest import SPEC_FILE, TEST_FILES_DIR


@pytest.mark.public_api
class TestSpecFile:
    def test_constructor(self, workdir):
        shutil.copy(os.path.join(TEST_FILES_DIR, SPEC_FILE), workdir)
        # check that no mandatory argument was added
        spec = SpecFile(SPEC_FILE)
        assert isinstance(spec, SpecFile)
        # check that arguments were not renamed or removed
        spec = SpecFile(path=SPEC_FILE, sources_location=workdir, predefined_macros=None)
        assert isinstance(spec, SpecFile)

    def test_update_changelog(self, spec_object):
        assert spec_object.update_changelog('test2') is None
        assert spec_object.update_changelog(changelog_entry='test') is None

    def test_set_version(self, spec_object):
        assert spec_object.set_version('1.2.3.4') is None
        assert spec_object.set_version(version='1.2.3') is None

    def test_get_version(self, spec_object):
        assert isinstance(spec_object.get_version(), str)

    def test_set_release_number(self, spec_object):
        assert spec_object.set_release_number('2') is None
        assert spec_object.set_release_number(release='1') is None

    def test_set_release(self, spec_object):
        assert spec_object.set_release('2') is None
        assert spec_object.set_release(release='1') is None

    def test_get_sources(self, spec_object):
        assert isinstance(spec_object.get_sources(), list)

    def test_tag(self, spec_object):
        assert isinstance(spec_object.tag(name='Source*', section='%package'), Tag)
        assert spec_object.tag(name='NotATag') is None

    def test_set_tag(self, spec_object):
        assert spec_object.set_tag('Version', '1.3.5') is None
        assert spec_object.set_tag(tag='Version', value='1.3.5', preserve_macros=True) is None

    def test_save(self, spec_object):
        assert spec_object.save() is None

    def test_udpate(self, spec_object):
        assert spec_object.update() is None



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import re
from typing import cast, Optional, Tuple

import git  # type: ignore
import requests

from rebasehelper.exceptions import RebaseHelperError
from rebasehelper.logger import CustomLogger


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class BugzillaHelper:

    """Class for working with Upstream Release Monitoring on bugzilla."""

    DIST_GIT_REPO_URL = 'https://src.fedoraproject.org/rpms'
    BUGZILLA_REST_API_URL = 'https://bugzilla.redhat.com/rest'
    UPSTREAM_RELEASE_MONITORING_USERNAME = 'upstream-release-monitoring'

    @classmethod
    def get_bugzilla_component(cls, bugzilla_id: str) -> str:
        """Gets a component of the bugzilla.

        Args:
            bugzilla_id: ID of the bugzilla.

        Returns:
            Component of the bugzilla.

        Raises:
            RebaseHelperError: If no such bugzilla exists or if the bugzilla
            was not created by Upstream Release Monitoring.

        """
        r = requests.get('{}/bug/{}'.format(cls.BUGZILLA_REST_API_URL, bugzilla_id))
        if not r.ok:
            raise RebaseHelperError('Could not obtain data from bugzilla')

        response_json = r.json()
        if 'error' in response_json:
            logger.error('Bugzilla error: %s', response_json['error'])
            raise RebaseHelperError('Could not obtain data from bugzilla')

        bug = response_json['bugs'][0]
        if bug['creator_detail']['email'] != cls.UPSTREAM_RELEASE_MONITORING_USERNAME:
            raise RebaseHelperError('The given bugzilla was not created by Upstream Release Monitoring')
        return bug['component'][0]

    @classmethod
    def get_version_from_comments(cls, bugzilla_id: str) -> Optional[str]:
        """Gets version from bugzilla comments.

        Args:
            bugzilla_id: ID of the bugzilla.

        Returns:
            Version specified by Upstream Release Monitoring in comments
            or None, if no version could be found.

        Raises:
            RebaseHelperError: If no such bugzilla exists.

        """
        r = requests.get('{}/bug/{}/comment'.format(cls.BUGZILLA_REST_API_URL, bugzilla_id))
        if not r.ok:
            raise RebaseHelperError('Could not obtain data from bugzilla')
        version = None
        comments = r.json()['bugs'][bugzilla_id]['comments']
        pattern = re.compile(r'^Latest upstream release: (?P<version>.*)\n')
        for comment in comments:
            if comment['creator'] != cls.UPSTREAM_RELEASE_MONITORING_USERNAME:
                continue
            match = pattern.match(comment['text'])
            if match:
                version = match.group('version')

        return version

    @classmethod
    def clone_repository(cls, component: str, bugzilla_id: str) -> str:
        """Clones remote dist-git repository of a component.

        Args:
            component: Package to clone.
            bugzilla_id: ID of the bugzilla.

        Returns:
            Path to the cloned repository.

        Raises:
            RebaseHelperError: If the directory, that the repository
            is supposed to be cloned into, exists.

        """
        path = os.path.abspath('{}-{}'.format(bugzilla_id, component))
        if os.path.exists(path):
            raise RebaseHelperError('Could not clone the repository because the directory '
                                    '{} already exists'.format(path))

        url = '{}/{}.git'.format(cls.DIST_GIT_REPO_URL, component)
        logger.info("Cloning %s into %s", url, path)
        git.Repo.clone_from(url, path)
        return path

    @classmethod
    def prepare_rebase_repository(cls, bugzilla_id: str) -> Tuple[str, str]:
        """Clones a repository based on Upstream Release Monitoring bugzilla.

        Args:
            bugzilla_id: ID of the bugzilla.

        Returns:
            Path of the cloned repository and version to rebase to.

        Raises:
            RebaseHelperError: If there was an error while obtaining
            data from bugzilla or if there was a problem while cloning
            the repository.

        """
        component = cls.get_bugzilla_component(bugzilla_id)
        version = cls.get_version_from_comments(bugzilla_id)
        if not version:
            raise RebaseHelperError('Could not obtain version from the bugzilla')

        path = cls.clone_repository(component, bugzilla_id)
        return path, version



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import shutil
from typing import cast

import pam  # type: ignore

from rebasehelper.helpers.path_helper import PathHelper
from rebasehelper.logger import CustomLogger
from rebasehelper.temporary_environment import TemporaryEnvironment


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


def get_mock_logfile_path(ret, results_dir, tmp_path=None):
    """
    Get path to logfile containing the error message

    :param ret: return code from mock
    :param results_dir: directory where logs will be stored
    :param tmp_path: temporary directory where logs are during build
    :return:
    """
    tmp_build_log_path = os.path.join(results_dir, 'build.log')
    tmp_mock_log_path = os.path.join(results_dir, 'mock_output.log')

    if tmp_path:
        # The logs are still located in the temporary build directory
        tmp_build_log_path = os.path.join(tmp_path, 'build.log')
        tmp_mock_log_path = os.path.join(tmp_path, 'mock_output.log')

    build_log_path = os.path.join(results_dir, 'build.log')
    mock_log_path = os.path.join(results_dir, 'mock_output.log')
    root_log_path = os.path.join(results_dir, 'root.log')

    # Mock return code classification based on https://pagure.io/koji/blob/c496bf9/f/builder/kojid#_481
    if ret == 1:
        if not os.path.exists(tmp_build_log_path) and os.path.exists(tmp_mock_log_path):
            logfile = mock_log_path
        else:
            logfile = build_log_path
    elif ret < 0:
        # koji error
        logfile = None
    else:
        logfile = root_log_path
    return logfile


def check_mock_privileges() -> bool:
    # try to authenticate as superuser using mock PAM service
    return pam.pam().authenticate('root', '', service='mock')


class BuildTemporaryEnvironment(TemporaryEnvironment):
    """Class representing temporary environment."""

    TEMPDIR_SOURCES: str = TemporaryEnvironment.TEMPDIR + '_SOURCES'
    TEMPDIR_SPEC: str = TemporaryEnvironment.TEMPDIR + '_SPEC'
    TEMPDIR_SPECS: str = TemporaryEnvironment.TEMPDIR + '_SPECS'
    TEMPDIR_RESULTS: str = TemporaryEnvironment.TEMPDIR + '_RESULTS'

    def __init__(self, sources, patches, spec, results_dir):
        super(BuildTemporaryEnvironment, self).__init__(self._build_env_exit_callback)
        self._env['results_dir'] = results_dir
        self.sources = sources
        self.patches = patches
        self.spec = spec

    def __enter__(self):
        obj = super(BuildTemporaryEnvironment, self).__enter__()
        log_message = "Copying '%s' to '%s'"
        # create the directory structure
        self._create_directory_structure()
        # copy sources
        for source in self.sources:
            logger.debug(log_message, source, self._env[self.TEMPDIR_SOURCES])
            shutil.copy(source, self._env[self.TEMPDIR_SOURCES])
        # copy patches
        for patch in self.patches:
            logger.debug(log_message, patch, self._env[self.TEMPDIR_SOURCES])
            shutil.copy(patch, self._env[self.TEMPDIR_SOURCES])
        # copy SPEC file
        spec_name = os.path.basename(self.spec)
        self._env[self.TEMPDIR_SPEC] = os.path.join(self._env[self.TEMPDIR_SPECS], spec_name)
        shutil.copy(self.spec, self._env[self.TEMPDIR_SPEC])
        logger.debug(log_message, self.spec, self._env[self.TEMPDIR_SPEC])

        return obj

    def _create_directory_structure(self):
        """Function creating the directory structure in the TemporaryEnvironment."""
        raise NotImplementedError('The create directory function has to be implemented in child class!')

    def _build_env_exit_callback(self, results_dir, **kwargs):
        """
        The function that is called just before the destruction of the TemporaryEnvironment.
        It copies packages and logs into the results directory.

        :param results_dir: absolute path to results directory
        :return:
        """
        log_message = "Copying '%s' '%s' to '%s'"
        # copy logs
        for log in PathHelper.find_all_files(kwargs[self.TEMPDIR_RESULTS], '*.log'):
            logger.debug(log_message, 'log', log, results_dir)
            shutil.copy(log, results_dir)
        # copy packages
        for package in PathHelper.find_all_files(kwargs[self.TEMPDIR], '*.rpm'):
            logger.debug(log_message, 'package', package, results_dir)
            shutil.copy(package, results_dir)


class RpmbuildTemporaryEnvironment(BuildTemporaryEnvironment):
    """Class representing temporary environment for RpmbuildBuildTool."""

    TEMPDIR_RPMBUILD: str = TemporaryEnvironment.TEMPDIR + '_RPMBUILD'
    TEMPDIR_BUILD: str = TemporaryEnvironment.TEMPDIR + '_BUILD'
    TEMPDIR_BUILDROOT: str = TemporaryEnvironment.TEMPDIR + '_BUILDROOT'
    TEMPDIR_RPMS: str = TemporaryEnvironment.TEMPDIR + '_RPMS'
    TEMPDIR_SRPMS: str = TemporaryEnvironment.TEMPDIR + '_SRPMS'

    def _create_directory_structure(self):
        # create rpmbuild directory structure
        for dir_name in ['RESULTS', 'rpmbuild']:
            self._env[self.TEMPDIR + '_' + dir_name.upper()] = os.path.join(
                self._env[self.TEMPDIR], dir_name)
            logger.debug("Creating '%s'",
                         self._env[self.TEMPDIR + '_' + dir_name.upper()])
            os.makedirs(self._env[self.TEMPDIR + '_' + dir_name.upper()])
        for dir_name in ['BUILD', 'BUILDROOT', 'RPMS', 'SOURCES', 'SPECS',
                         'SRPMS']:
            self._env[self.TEMPDIR + '_' + dir_name] = os.path.join(
                self._env[self.TEMPDIR_RPMBUILD],
                dir_name)
            logger.debug("Creating '%s'",
                         self._env[self.TEMPDIR + '_' + dir_name])
            os.makedirs(self._env[self.TEMPDIR + '_' + dir_name])


class MockTemporaryEnvironment(BuildTemporaryEnvironment):
    """Class representing temporary environment for MockBuildTool."""

    def _create_directory_structure(self):
        # create directory structure
        for dir_name in ['SOURCES', 'SPECS', 'RESULTS']:
            self._env[self.TEMPDIR + '_' + dir_name] = os.path.join(
                self._env[self.TEMPDIR], dir_name)
            logger.debug("Creating '%s'",
                         self._env[self.TEMPDIR + '_' + dir_name])
            os.makedirs(self._env[self.TEMPDIR + '_' + dir_name])



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import random
import string
import sys
import time

from typing import List, cast

from rebasehelper.exceptions import RebaseHelperError
from rebasehelper.helpers.console_helper import ConsoleHelper
from rebasehelper.helpers.rpm_helper import RpmHelper
from rebasehelper.helpers.download_helper import DownloadHelper
from rebasehelper.logger import CustomLogger

koji_helper_functional: bool
try:
    import koji  # type: ignore
    from koji_cli.lib import TaskWatcher  # type: ignore
except ImportError:
    koji_helper_functional = False
else:
    koji_helper_functional = True


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class KojiHelper:

    functional: bool = koji_helper_functional

    @classmethod
    def create_session(cls, login=False, profile='koji'):
        """Creates new Koji session and immediately logs in to a Koji hub.

        Args:
            login (bool): Whether to perform a login.
            profile (str): Koji profile to use.

        Returns:
            koji.ClientSession: Newly created session instance.

        Raises:
            RebaseHelperError: If login failed.

        """
        config = koji.read_config(profile)
        session = koji.ClientSession(config['server'], opts=config)
        if not login:
            return session
        try:
            session.gssapi_login()
        except Exception:  # pylint: disable=broad-except
            pass
        else:
            return session
        # fall back to kerberos login (doesn't work with python3)
        exc = (koji.AuthError, koji.krbV.Krb5Error) if koji.krbV else koji.AuthError
        try:
            session.krb_login()
        except exc as e:
            raise RebaseHelperError('Login failed: {}'.format(str(e))) from e
        else:
            return session

    @classmethod
    def upload_srpm(cls, session, srpm):
        """Uploads SRPM to a Koji hub.

        Args:
            session (koji.ClientSession): Active Koji session instance.
            srpm (str): Valid path to SRPM.

        Returns:
            str: Remote path to the uploaded SRPM.

        Raises:
            RebaseHelperError: If upload failed.

        """
        def progress(uploaded, total, chunksize, t1, t2):  # pylint: disable=unused-argument
            DownloadHelper.progress(total, uploaded, upload_start)
        suffix = ''.join(random.choice(string.ascii_letters) for _ in range(8))
        path = os.path.join('cli-build', str(time.time()), suffix)
        logger.info('Uploading SRPM')
        try:
            try:
                upload_start = time.time()
                session.uploadWrapper(srpm, path, callback=progress)
            except koji.GenericError as e:
                raise RebaseHelperError('Upload failed: {}'.format(str(e))) from e
        finally:
            sys.stdout.write('\n')
            sys.stdout.flush()
        return os.path.join(path, os.path.basename(srpm))

    @classmethod
    def get_task_url(cls, session, task_id):
        return '/'.join([session.opts['weburl'], 'taskinfo?taskID={}'.format(task_id)])

    @classmethod
    def display_task_results(cls, tasks):
        """Prints states of Koji tasks.

        Args:
            tasks (list): List of koji.TaskWatcher instances.

        """
        for task in [t for t in tasks if t.level == 0]:
            state = task.info['state']
            task_label = task.str()
            logger.info('State %s (%s)', state, task_label)
            if state == koji.TASK_STATES['CLOSED']:
                logger.info('%s completed successfully', task_label)
            elif state == koji.TASK_STATES['FAILED']:
                logger.info('%s failed', task_label)
            elif state == koji.TASK_STATES['CANCELED']:
                logger.info('%s was canceled', task_label)
            else:
                # shouldn't happen
                logger.info('%s has not completed', task_label)

    @classmethod
    def watch_koji_tasks(cls, session, tasklist):
        """Waits for Koji tasks to finish and prints their states.

        Args:
            session (koji.ClientSession): Active Koji session instance.
            tasklist (list): List of task IDs.

        Returns:
            dict: Dictionary mapping task IDs to their states or None if interrupted.

        """
        if not tasklist:
            return None
        sys.stdout.flush()
        rh_tasks = {}
        try:
            tasks = {}
            for task_id in tasklist:
                task_id = int(task_id)
                tasks[task_id] = TaskWatcher(task_id, session, quiet=False)
            while True:
                all_done = True
                for task_id, task in list(tasks.items()):
                    with ConsoleHelper.Capturer(stdout=True) as capturer:
                        changed = task.update()
                    for line in capturer.stdout.split('\n'):
                        if line:
                            logger.info(line)
                    info = session.getTaskInfo(task_id)
                    state = task.info['state']
                    if state == koji.TASK_STATES['FAILED']:
                        return {info['id']: state}
                    else:
                        # FIXME: multiple arches
                        if info['arch'] == 'x86_64' or info['arch'] == 'noarch':
                            rh_tasks[info['id']] = state
                    if not task.is_done():
                        all_done = False
                    else:
                        if changed:
                            # task is done and state just changed
                            cls.display_task_results(list(tasks.values()))
                        if not task.is_success():
                            rh_tasks = None
                    for child in session.getTaskChildren(task_id):
                        child_id = child['id']
                        if child_id not in list(tasks.keys()):
                            tasks[child_id] = TaskWatcher(child_id, session, task.level + 1, quiet=False)
                            with ConsoleHelper.Capturer(stdout=True) as capturer:
                                tasks[child_id].update()
                            for line in capturer.stdout.split('\n'):
                                if line:
                                    logger.info(line)
                            info = session.getTaskInfo(child_id)
                            state = task.info['state']
                            if state == koji.TASK_STATES['FAILED']:
                                return {info['id']: state}
                            else:
                                # FIXME: multiple arches
                                if info['arch'] == 'x86_64' or info['arch'] == 'noarch':
                                    rh_tasks[info['id']] = state
                            # If we found new children, go through the list again,
                            # in case they have children also
                            all_done = False
                if all_done:
                    cls.display_task_results(list(tasks.values()))
                    break
                sys.stdout.flush()
                time.sleep(1)
        except KeyboardInterrupt:
            rh_tasks = None
        return rh_tasks

    @classmethod
    def download_task_results(cls, session, tasklist, destination):
        """Downloads packages and logs of finished Koji tasks.

        Args:
            session (koji.ClientSession): Active Koji session instance.
            tasklist (list): List of task IDs.
            destination (str): Path where to download files to.

        Returns:
            tuple: List of downloaded RPMs and list of downloaded logs.

        Raises:
            DownloadError: If download failed.

        """
        rpms: List[str] = []
        logs: List[str] = []
        for task_id in tasklist:
            logger.info('Downloading packages and logs for task %s', task_id)
            task = session.getTaskInfo(task_id, request=True)
            if task['state'] in [koji.TASK_STATES['FREE'], koji.TASK_STATES['OPEN']]:
                logger.info('Task %s is still running!', task_id)
                continue
            elif task['state'] != koji.TASK_STATES['CLOSED']:
                logger.info('Task %s did not complete successfully!', task_id)
            if task['method'] == 'buildArch':
                tasks = [task]
            elif task['method'] == 'build':
                opts = dict(parent=task_id, method='buildArch', decode=True,
                            state=[koji.TASK_STATES['CLOSED'], koji.TASK_STATES['FAILED']])
                tasks = session.listTasks(opts=opts)
            else:
                logger.info('Task %s is not a build or buildArch task!', task_id)
                continue
            for task in tasks:
                base_path = koji.pathinfo.taskrelpath(task['id'])
                output = session.listTaskOutput(task['id'])
                for filename in output:
                    local_path = os.path.join(destination, filename)
                    download = False
                    fn, ext = os.path.splitext(filename)
                    if ext == '.rpm':
                        if task['state'] != koji.TASK_STATES['CLOSED']:
                            continue
                        if local_path not in rpms:
                            nevra = RpmHelper.split_nevra(fn)
                            # FIXME: multiple arches
                            download = nevra['arch'] in ['noarch', 'x86_64']
                            if download:
                                rpms.append(local_path)
                    else:
                        if local_path not in logs:
                            download = True
                            logs.append(local_path)
                    if download:
                        logger.info('Downloading file %s', filename)
                        url = '/'.join([session.opts['topurl'], 'work', base_path, filename])
                        DownloadHelper.download_file(url, local_path)
        return rpms, logs

    @classmethod
    def get_latest_build(cls, session, package):
        """Looks up latest Koji build of a package.

        Args:
            session (koji.ClientSession): Active Koji session instance.
            package (str): Package name.

        Returns:
            tuple: Found latest package version and Koji build ID.

        """
        builds = session.getLatestBuilds('rawhide', package=package)
        if builds:
            build = builds.pop()
            return build['version'], build['id']
        return None, None

    @classmethod
    def get_build(cls, session, package, version):
        """Looks up Koji build of a specific version of a package.

        Args:
            session (koji.ClientSession): Active Koji session instance.
            package (str): Package name.
            version (str): Package version.

        Returns:
            tuple: Found latest package version and Koji build ID.

        """
        builds = session.listTagged('rawhide', inherit=True, package=package)
        if builds:
            for build in builds:
                if build['version'] == version:
                    return build['version'], build['id']
        return None, None

    @classmethod
    def download_build(cls, session, build_id, destination, arches):
        """Downloads RPMs and logs of a Koji build.

        Args:
            session (koji.ClientSession): Active Koji session instance.
            build_id (str): Koji build ID.
            destination (str): Path where to download files to.
            arches (list): List of architectures to be downloaded.

        Returns:
            tuple: List of downloaded RPMs and list of downloaded logs.

        Raises:
            DownloadError: If download failed.

        """
        build = session.getBuild(build_id)
        pathinfo = koji.PathInfo(topdir=session.opts['topurl'])
        rpms: List[str] = []
        logs: List[str] = []
        os.makedirs(destination, exist_ok=True)
        for pkg in session.listBuildRPMs(build_id):
            if pkg['arch'] not in arches:
                continue
            rpmpath = pathinfo.rpm(pkg)
            local_path = os.path.join(destination, os.path.basename(rpmpath))
            if local_path not in rpms:
                url = pathinfo.build(build) + '/' + rpmpath
                DownloadHelper.download_file(url, local_path)
                rpms.append(local_path)
        for logfile in session.getBuildLogs(build_id):
            if logfile['dir'] not in arches:
                continue
            local_path = os.path.join(destination, logfile['name'])
            if local_path not in logs:
                url = pathinfo.topdir + '/' + logfile['path']
                DownloadHelper.download_file(url, local_path)
                logs.append(local_path)
        return rpms, logs

    @classmethod
    def get_old_build_info(cls, package_name, package_version):
        """Gets old build info from Koji.

        Args:
            package_name (str): Package name from specfile.
            package_version (str): Package version from specfile.

        Returns:
            tuple: Koji build id, package version.

        """
        if cls.functional:
            session = KojiHelper.create_session()
            koji_version, koji_build_id = KojiHelper.get_build(session, package_name, package_version)
            if koji_version:
                return koji_build_id, koji_version
            else:
                logger.warning('Unable to find old version Koji build!')
                return None, None
        else:
            logger.warning('Unable to get old version Koji build!')
            return None, None



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
from typing import cast

from rebasehelper.helpers.copr_helper import CoprHelper
from rebasehelper.types import Options
from rebasehelper.logger import CustomLogger
from rebasehelper.exceptions import RebaseHelperError
from rebasehelper.plugins.build_tools.rpm import BuildToolBase
from rebasehelper.exceptions import BinaryPackageBuildError


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class Copr(BuildToolBase):
    """
    Class representing Copr build tool.
    """

    OPTIONS: Options = [
        {
            'name': ['--copr-project-permanent'],
            'switch': True,
            'default': False,
            'help': 'make the created copr project permanent'
        },
        {
            'name': ['--copr-project-frontpage'],
            'switch': True,
            'default': False,
            'help': 'make the created copr project visible on the frontpage',
        },
        {
            'name': ['--copr-chroots'],
            'type': lambda s: s.split(','),
            'default': ['fedora-rawhide-x86_64'],
            'help': 'comma-separated list of chroots to create copr project with',
        },
    ]

    CREATES_TASKS: bool = True

    PREFIX: str = 'rebase-helper-'
    DESCRIPTION: str = 'Repository containing rebase-helper builds.'
    INSTRUCTIONS: str = '''You can use this repository to test functionality
                         of rebased packages.'''

    @classmethod
    def _build_rpms(cls, srpm, name, **kwargs):
        project = cls.PREFIX + name
        client = CoprHelper.get_client()
        options = kwargs.get('app_kwargs', {})
        hide = not options.get('copr_project_frontpage')
        permanent = options.get('copr_project_permanent')
        chroots = options.get('copr_chroots')

        CoprHelper.create_project(client, project, chroots, cls.DESCRIPTION, cls.INSTRUCTIONS, permanent, hide)
        build_id = CoprHelper.build(client, project, srpm)
        if kwargs['builds_nowait']:
            return None, None, build_id
        build_url = CoprHelper.get_build_url(client, build_id)
        logger.info('Copr build is here: %s', build_url)
        failed = not CoprHelper.watch_build(client, build_id)
        destination = os.path.dirname(srpm).replace('SRPM', 'RPM')
        rpms, logs = CoprHelper.download_build(client, build_id, destination)
        logs.append(build_url)
        if failed:
            logger.error('Copr build failed %s', build_url)
            raise BinaryPackageBuildError(logs=logs)
        return rpms, logs, build_id

    @classmethod
    def build(cls, spec, results_dir, srpm, **kwargs):
        """
        Builds the RPMs using copr

        :param spec: SpecFile object
        :param results_dir: absolute path to DIR where results should be stored
        :param srpm: absolute path to SRPM
        :return: dict with:
                 'rpm' -> list with absolute paths to RPMs
                 'logs' -> list with absolute paths to build_logs
                 'copr_build_id' -> ID of copr build
        """
        rpms, logs, build_id = cls._build_rpms(srpm, **kwargs)
        return dict(rpm=rpms, logs=logs, copr_build_id=build_id)

    @classmethod
    def get_task_info(cls, build_dict):
        client = CoprHelper.get_client()
        build_url = CoprHelper.get_build_url(client, build_dict['copr_build_id'])
        message = "Copr build for '%s' version is: %s"
        return message % (build_dict['version'], build_url)

    @classmethod
    def get_detached_task(cls, task_id, results_dir):
        client = CoprHelper.get_client()
        build_id = int(task_id)
        status = CoprHelper.get_build_status(client, build_id)
        if status in ['importing', 'pending', 'starting', 'running']:
            raise RebaseHelperError('Copr build is not finished yet. Try again later')
        else:
            rpm, logs = CoprHelper.download_build(client, build_id, results_dir)
            if status not in ['succeeded', 'skipped']:
                logger.info('Copr build %d did not complete successfully', build_id)
                raise BinaryPackageBuildError(logs=logs)
            return rpm, logs

    @classmethod
    def wait_for_task(cls, build_dict, task_id, results_dir):
        client = CoprHelper.get_client()
        build_id = int(task_id)
        failed = not CoprHelper.watch_build(client, build_id)
        rpms, logs = CoprHelper.download_build(client, build_id, results_dir)
        build_url = CoprHelper.get_build_url(client, build_id)
        logs.append(build_url)
        if failed:
            logger.error('Copr build failed %s', build_url)
            raise BinaryPackageBuildError(logs=logs)
        return rpms, logs



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

from textwrap import dedent
from types import SimpleNamespace

import pytest  # type: ignore

from rebasehelper.tags import Tags
from rebasehelper.plugins.spec_hooks.typo_fix import TypoFix
from rebasehelper.plugins.spec_hooks.pypi_url_fix import PyPIURLFix
from rebasehelper.plugins.spec_hooks.escape_macros import EscapeMacros
from rebasehelper.plugins.spec_hooks.replace_old_version import ReplaceOldVersion
from rebasehelper.plugins.spec_hooks.paths_to_rpm_macros import PathsToRPMMacros


class TestSpecHook:

    @pytest.mark.parametrize('spec_attributes', [
        {
            'spec_content': '%changelog\n- This is chnagelog entry with some indentional typos'
        }
    ])
    def test_typo_fix_spec_hook(self, mocked_spec_object):
        assert '- This is chnagelog entry with some indentional typos' in \
               mocked_spec_object.spec_content.section('%changelog')
        TypoFix.run(mocked_spec_object, mocked_spec_object)
        assert '- This is changelog entry with some intentional typos' in \
               mocked_spec_object.spec_content.section('%changelog')

    @pytest.mark.parametrize('spec_attributes', [
        {
            'spec_content': dedent("""\
                %files
                /usr/share/man/man1/*
                /usr/bin/%{name}
                %config(noreplace) /etc/test/test.conf

                %files devel
                %{_bindir}/test_example
                %{_libdir}/my_test.so
                /usr/share/test1.txt
                /no/macros/here
                """),
            'macros': {},
        }
    ])
    def test_paths_to_rpm_macros_spec_hook(self, mocked_spec_object):
        files = [
            '%{_mandir}/man1/*',
            '%{_bindir}/%{name}',
            '%config(noreplace) %{_sysconfdir}/test/test.conf',
            '',
        ]
        files_devel = [
            '%{_bindir}/test_example',
            '%{_libdir}/my_test.so',
            '%{_datadir}/test1.txt',
            '/no/macros/here',
        ]
        PathsToRPMMacros.run(mocked_spec_object, mocked_spec_object)
        assert files == mocked_spec_object.spec_content.section('%files')
        assert files_devel == mocked_spec_object.spec_content.section('%files devel')

    @pytest.mark.parametrize('spec_attributes', [
        {
            'spec_content': dedent("""\
                Source9: https://test.com/#/1.0/%{name}-hardcoded-version-1.0.2.tar.gz

                %build
                autoreconf -vi # Unescaped macros %name %{name}
                """)
        }
    ])
    def test_escape_macros_spec_hook(self, mocked_spec_object):
        EscapeMacros.run(mocked_spec_object, mocked_spec_object)
        build = mocked_spec_object.spec_content.section('%build')
        assert build[0] == "autoreconf -vi # Unescaped macros %%name %%{name}"
        # Test that the string after `#` wasn't recognized as a comment.
        source9 = mocked_spec_object.get_raw_tag_value('Source9')
        assert source9 == "https://test.com/#/1.0/%{name}-hardcoded-version-1.0.2.tar.gz"

    @pytest.mark.parametrize('replace_with_macro', [
        True,
        False,
    ], ids=[
        'new-version',
        'macro',
    ])
    @pytest.mark.parametrize('spec_attributes', [
        {
            'spec_content': dedent("""\
                Version: 1.0.2
                Source9: https://test.com/#/1.0/%{name}-hardcoded-version-1.0.2b1.tar.gz
                Recommends: test > 1.0.2

                %changelog
                * Wed Apr 26 2017 Nikola Forró <nforro@redhat.com> - 1.0.2-34
                - Update to 1.0.2
                """),
            'header': SimpleNamespace(version='1.0.2', release='0.1.b1'),
        }
    ])
    def test_replace_old_version_spec_hook(self, mocked_spec_object, mocked_spec_object_copy, replace_with_macro):
        mocked_spec_object_copy.header.version = '1.1.0'
        mocked_spec_object_copy.header.release = '1'
        mocked_spec_object_copy.set_raw_tag_value('Version', '1.1.0')
        ReplaceOldVersion.run(mocked_spec_object, mocked_spec_object_copy,
                              replace_old_version_with_macro=replace_with_macro)
        # The spec is not saved due to mocking, refresh tags for the assertions
        mocked_spec_object_copy.tags = Tags(mocked_spec_object_copy.spec_content, mocked_spec_object_copy.spec_content)

        # Check if the version has been updated
        test_source = mocked_spec_object_copy.get_raw_tag_value('Source9')
        if replace_with_macro:
            assert test_source == 'https://test.com/#/1.1/%{name}-hardcoded-version-%{version}.tar.gz'
        else:
            assert test_source == 'https://test.com/#/1.1/%{name}-hardcoded-version-1.1.0.tar.gz'

        # Check if dependency and Version tags are ignored
        assert mocked_spec_object_copy.get_raw_tag_value('Recommends') == 'test > 1.0.2'
        assert mocked_spec_object_copy.get_raw_tag_value('Version') == '1.1.0'

        # Check if version in changelog hasn't been changed
        changelog = mocked_spec_object_copy.spec_content.section('%changelog')
        assert '1.0.2' in changelog[0]

    @pytest.mark.parametrize('spec_attributes', [
        {
            'spec_content': dedent("""\
                URL: https://pypi.python.org/pypi/%{name}
                Source0: https://pypi.python.org/.../%{name}.%{version}.tar.gz
                """)
        }
    ])
    def test_pypi_to_python_hosted_url_trans(self, mocked_spec_object):
        PyPIURLFix.run(mocked_spec_object, mocked_spec_object)
        expected = {
            'URL': 'https://pypi.org/project/%{name}',
            'Source0': 'https://files.pythonhosted.org/.../%{name}.%{version}.tar.gz',
        }
        for tag, value in expected.items():
            assert value == mocked_spec_object.get_raw_tag_value(tag)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import bz2
import logging
import lzma
import os
import shutil
import tarfile
import zipfile
from typing import Dict, Type, cast

from rebasehelper.logger import CustomLogger
from rebasehelper.helpers.path_helper import PathHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


# supported archive types
archive_types: Dict[str, Type['ArchiveTypeBase']] = {}


def register_archive_type(archive):
    archive_types[archive.EXTENSION] = archive
    return archive


class ArchiveTypeBase:
    """Base class for various archive types"""

    EXTENSION: str = ''

    @classmethod
    def match(cls, filename):
        """
        Checks if the filename matches the archive type. If yes, returns
        True, otherwise returns False.
        """
        return filename.endswith(cls.EXTENSION)

    @classmethod
    def open(cls, filename):
        """
        Opens archive with the given filename and returns the proper
        archive type object.
        """
        raise NotImplementedError()

    @classmethod
    def extract(cls, archive, filename, path):
        """
        Extracts the archive into the given path

        :param path: Path where to extract the archive to.
        :return:
        """
        raise NotImplementedError()


@register_archive_type
class TarXzArchiveType(ArchiveTypeBase):
    EXTENSION: str = '.tar.xz'

    @classmethod
    def open(cls, filename):
        return tarfile.open(mode='r', fileobj=lzma.LZMAFile(filename, 'r'))

    @classmethod
    def extract(cls, archive, filename, path):
        archive.extractall(path)


@register_archive_type
class TarBz2ArchiveType(ArchiveTypeBase):
    EXTENSION: str = '.tar.bz2'

    @classmethod
    def open(cls, filename):
        if filename.endswith(TarBz2ArchiveType.EXTENSION):
            return tarfile.TarFile.open(filename)
        else:
            return bz2.BZ2File(filename)

    @classmethod
    def extract(cls, archive, filename, path):
        if filename.endswith(TarBz2ArchiveType.EXTENSION):
            archive.extractall(path)
        else:
            data = archive.read()
            if not os.path.exists(path):
                os.mkdir(path)
            with open(os.path.join(path, os.path.basename(filename[:-len(cls.EXTENSION)])), 'wb') as f:
                f.write(data)


@register_archive_type
class Bz2ArchiveType(TarBz2ArchiveType):
    EXTENSION: str = '.bz2'


@register_archive_type
class TarGzArchiveType(TarBz2ArchiveType):
    EXTENSION: str = '.tar.gz'

    @classmethod
    def open(cls, filename):
        return tarfile.TarFile.open(filename)

    @classmethod
    def extract(cls, archive, filename, path):
        archive.extractall(path)


@register_archive_type
class TgzArchiveType(TarGzArchiveType):
    EXTENSION: str = '.tgz'


@register_archive_type
class TarArchiveType(TarGzArchiveType):
    EXTENSION: str = '.tar'


@register_archive_type
class CrateArchiveType(TarGzArchiveType):
    EXTENSION: str = '.crate'


@register_archive_type
class ZipArchiveType(ArchiveTypeBase):
    EXTENSION: str = '.zip'

    @classmethod
    def match(cls, filename):
        return zipfile.is_zipfile(filename)

    @classmethod
    def open(cls, filename):
        return zipfile.ZipFile(filename, 'r')

    @classmethod
    def extract(cls, archive, filename, path):
        archive.extractall(path)


@register_archive_type
class GemArchiveType(ArchiveTypeBase):
    EXTENSION: str = '.gem'

    class GemArchive:
        def __init__(self, filename):
            self.tmp = PathHelper.get_temp_dir()
            TarArchiveType.extract(TarArchiveType.open(filename), '', self.tmp)
            self.data = TarGzArchiveType.open(os.path.join(self.tmp, 'data.tar.gz'))

        def extract(self, path):
            TarGzArchiveType.extract(self.data, '', path)

        def close(self):
            self.data.close()
            shutil.rmtree(self.tmp, onerror=lambda func, path, excinfo: shutil.rmtree(path))

    @classmethod
    def open(cls, filename):
        return cls.GemArchive(filename)

    @classmethod
    def extract(cls, archive, filename, path):
        tld = os.path.join(path, os.path.basename(filename[:-len(cls.EXTENSION)]))
        os.makedirs(tld)
        archive.extract(tld)


class Archive:
    """Class representing an archive with sources"""

    def __init__(self, filename):
        self._filename = filename
        self._archive_type = None

        for archive_type in archive_types.values():
            if archive_type.match(self._filename):
                self._archive_type = archive_type

        if self._archive_type is None:
            raise NotImplementedError('Unsupported archive type')

    def extract_archive(self, path):
        """
        Extracts the archive into the given path

        :param path: Path where to extract the archive to.
        """
        logger.verbose('Extracting %s to %s', self._filename, path)

        try:
            archive = self._archive_type.open(self._filename)
        except (EOFError, tarfile.ReadError, lzma.LZMAError) as e:
            raise IOError(str(e)) from e

        try:
            self._archive_type.extract(archive, self._filename, path)
        finally:
            archive.close()

    @classmethod
    def get_supported_archives(cls):
        return archive_types.keys()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
from typing import cast

import git  # type: ignore

from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.logger import CustomLogger


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class GitHelper:

    """Class which operates with git repositories"""

    # provide fallback values if system is not configured
    GIT_USER_NAME: str = 'rebase-helper'
    GIT_USER_EMAIL: str = 'rebase-helper@localhost.local'

    @classmethod
    def get_user(cls):
        try:
            return git.cmd.Git().config('user.name', get=True, stdout_as_string=True)
        except git.GitCommandError:
            logger.warning("Failed to get configured git user name, using '%s'", cls.GIT_USER_NAME)
            return cls.GIT_USER_NAME

    @classmethod
    def get_email(cls):
        try:
            return git.cmd.Git().config('user.email', get=True, stdout_as_string=True)
        except git.GitCommandError:
            logger.warning("Failed to get configured git user email, using '%s'", cls.GIT_USER_EMAIL)
            return cls.GIT_USER_EMAIL

    @classmethod
    def run_mergetool(cls, repo):
        # we can't use GitPython here, as it doesn't allow
        # for the command to attach to stdout directly
        cwd = os.getcwd()
        try:
            os.chdir(repo.working_tree_dir)
            ProcessHelper.run_subprocess(['git', 'mergetool'])
        finally:
            os.chdir(cwd)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os

import pytest  # type: ignore

from rebasehelper.helpers.path_helper import PathHelper


class TestPathHelper:
    """ PathHelper tests """

    @pytest.fixture
    def filelist(self):
        files = [
            'file',
            'ffile',
            'ppythooon',
            'dir1/fileee',
            'dir1/faa/pythooon',
            'dir1/foo/pythooon',
            'dir1/foo/bar/file',
            'dir1/foo/baz/file',
            'dir1/baz/ffile',
            'dir1/bar/file',
            'dir1/baz/bar/ffile',
            'dir1/baz/bar/test.spec',
        ]

        for f in files:
            try:
                os.makedirs(os.path.dirname(f))
            except OSError:
                pass
            with open(f, 'w') as fd:
                fd.write(f)

        return files

    class TestFindFirstDirWithFile:
        """ PathHelper - find_first_dir_with_file() tests """
        def test_find_file(self, filelist):
            assert PathHelper.find_first_dir_with_file(
                "dir1", "file") == os.path.abspath(
                os.path.dirname(filelist[9]))
            assert PathHelper.find_first_dir_with_file(
                os.path.curdir, "file") == os.path.abspath(os.path.dirname(filelist[0]))
            assert PathHelper.find_first_dir_with_file(
                "dir1/baz", "file") is None

        def test_find_ffile(self, filelist):
            assert PathHelper.find_first_dir_with_file(
                "dir1", "*le") == os.path.abspath(
                os.path.dirname(filelist[9]))
            assert PathHelper.find_first_dir_with_file(
                "dir1", "ff*") == os.path.abspath(
                os.path.dirname(filelist[8]))
            assert PathHelper.find_first_dir_with_file(
                "dir1/foo", "ff*") is None

        def test_find_pythoon(self, filelist):
            assert PathHelper.find_first_dir_with_file(
                "dir1", "pythooon") == os.path.abspath(
                os.path.dirname(filelist[4]))
            assert PathHelper.find_first_dir_with_file(
                os.path.curdir, "py*n") == os.path.abspath(os.path.dirname(filelist[4]))
            assert PathHelper.find_first_dir_with_file(
                "dir1/bar", "pythooon") is None

    class TestFindFirstFile:
        """ PathHelper - find_first_file() tests """
        def test_find_file(self, filelist):
            assert PathHelper.find_first_file(
                "dir1", "file") == os.path.abspath(filelist[9])
            assert PathHelper.find_first_file(
                os.path.curdir, "file") == os.path.abspath(filelist[0])
            assert PathHelper.find_first_file("dir1/baz", "file") is None

        def test_find_ffile(self, filelist):
            assert PathHelper.find_first_file(
                "dir1", "*le") == os.path.abspath(filelist[9])
            assert PathHelper.find_first_file(
                "dir1", "ff*") == os.path.abspath(filelist[8])
            assert PathHelper.find_first_file("dir1/foo", "ff*") is None

        def test_find_pythoon(self, filelist):
            assert PathHelper.find_first_file(
                "dir1", "pythooon") == os.path.abspath(filelist[4])
            assert PathHelper.find_first_file(
                os.path.curdir, "py*n") == os.path.abspath(filelist[4])
            assert PathHelper.find_first_file("dir1/bar", "pythooon") is None

        def test_find_with_recursion(self, filelist):
            assert PathHelper.find_first_file(os.path.curdir, "*.spec", 0) is None
            assert PathHelper.find_first_file(os.path.curdir, "*.spec", 1) is None
            assert PathHelper.find_first_file(os.path.curdir, "*.spec", 2) is None
            assert PathHelper.find_first_file(os.path.curdir, "*.spec", 3) is None
            assert PathHelper.find_first_file(os.path.curdir, "*.spec", 4) == os.path.abspath(filelist[-1])

        def test_find_without_recursion(self, filelist):
            assert PathHelper.find_first_file(os.path.curdir, "*.spec") == os.path.abspath(filelist[-1])



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import io
import logging
import os
import re
from typing import Dict, List, Optional, cast

from rebasehelper.exceptions import RebaseHelperError, CheckerNotFoundError
from rebasehelper.logger import CustomLogger
from rebasehelper.results_store import results_store
from rebasehelper.plugins.checkers import BaseChecker, CheckerCategory
from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.helpers.rpm_helper import RpmHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class RpmDiff(BaseChecker):
    """rpmdiff compare tool"""

    DEFAULT: bool = True
    CATEGORY: Optional[CheckerCategory] = CheckerCategory.RPM

    CMD: str = 'rpmdiff'
    CHECKER_TAGS: List[str] = ['added', 'removed', 'changed', 'moved', 'renamed']

    @classmethod
    def is_available(cls):
        try:
            return ProcessHelper.run_subprocess([cls.CMD, '--help'], output_file=ProcessHelper.DEV_NULL) == 0
        except (IOError, OSError):
            return False

    @classmethod
    def _get_rpms(cls, rpm_list):
        rpm_dict = {}
        for rpm_name in rpm_list:
            rpm_dict[RpmHelper.get_header_from_rpm(rpm_name).name] = rpm_name
        return rpm_dict

    @classmethod
    def _unpack_rpm(cls, rpm_name):
        pass

    @classmethod
    def _analyze_logs(cls, output, results_dict):
        removed_things = ['.build-id', '.dwz', 'PROVIDE', 'REQUIRES']
        for line in output:
            if [x for x in removed_things if x in line]:
                continue

            fields = line.strip().split()
            logger.verbose(fields)
            if line.startswith('removed'):
                results_dict['removed'].append(fields[1])
                continue
            if line.startswith('added'):
                results_dict['added'].append(fields[1])
                continue

            if re.match(r'(S..|..5)........', fields[0]):
                # size or checksum changed
                results_dict['changed'].append(fields[1])
        return results_dict

    @classmethod
    def update_added_removed(cls, results_dict):
        added = []
        removed = []
        for item in results_dict['removed']:
            found = [x for x in results_dict['added'] if os.path.basename(item) == os.path.basename(x)]
            if not found:
                removed.append(item)

        for item in results_dict['added']:
            found = [x for x in results_dict['removed'] if os.path.basename(item) == os.path.basename(x)]
            if not found:
                added.append(item)
        results_dict['added'] = added
        results_dict['removed'] = removed
        return results_dict

    @classmethod
    def run_check(cls, results_dir, **kwargs):
        """Compares old and new RPMs using rpmdiff"""
        results_dict: Dict[str, List[str]] = {}

        for tag in cls.CHECKER_TAGS:
            results_dict[tag] = []

        cls.results_dir = os.path.join(results_dir, cls.name)
        cls.prepare_results_dir()

        # Only S (size), M(mode) and 5 (checksum) are now important
        not_catched_flags = ['T', 'F', 'G', 'U', 'V', 'L', 'D', 'N']
        old_pkgs = cls._get_rpms(results_store.get_old_build().get('rpm', None))
        new_pkgs = cls._get_rpms(results_store.get_new_build().get('rpm', None))
        for key, value in old_pkgs.items():
            if 'debuginfo' in key or 'debugsource' in key:
                # skip debug{info,source} packages
                continue
            cmd = [cls.CMD]
            # TODO modify to online command
            for x in not_catched_flags:
                cmd.extend(['-i', x])
            cmd.append(value)
            # We would like to build correct old package against correct new packages
            try:
                cmd.append(new_pkgs[key])
            except KeyError:
                logger.warning('New version of package %s was not found!', key)
                continue
            output = io.StringIO()
            try:
                ProcessHelper.run_subprocess(cmd, output_file=output)
            except OSError as e:
                raise CheckerNotFoundError("Checker '{}' was not found or installed.".format(cls.name)) from e
            results_dict = cls._analyze_logs(output, results_dict)
        results_dict = cls.update_added_removed(results_dict)
        cls.results_dict = {k: v for k, v in results_dict.items() if v}
        lines: List[str] = []
        for key, val in results_dict.items():
            if val:
                if lines:
                    lines.append('')
                lines.append('Following files were {}:'.format(key))
                lines.extend(val)

        rpmdiff_report = os.path.join(cls.results_dir, 'report.txt')

        counts = {k: len(v) for k, v in results_dict.items()}

        try:
            with open(rpmdiff_report, "w") as f:
                f.write('\n'.join(lines))
        except IOError as e:
            raise RebaseHelperError("Unable to write result from {} to '{}'".format(cls.name, rpmdiff_report)) from e

        return {'path': cls.get_checker_output_dir_short(), 'files_changes': counts}

    @classmethod
    def format(cls, data):
        """
        Format rpmdiff output for outputtool
        :param data: rpmdiff output data dictionary
        :return: formated rpmdiff output list of strings
        """
        output_lines = [cls.get_underlined_title("rpmdiff")]

        output_lines.append(" - {} added files".format(data['files_changes']['added']))
        output_lines.append(" - {} changed files".format(data['files_changes']['changed']))
        output_lines.append(" - {} removed files".format(data['files_changes']['removed']))

        output_lines.append("Details in {}:".format(data['path']))
        output_lines.append(" - report.txt")

        return output_lines



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import re
from typing import List, Tuple

from rebasehelper.plugins.spec_hooks import BaseSpecHook
from rebasehelper.types import PackageCategories
from rebasehelper.specfile import PackageCategory


class PyPIURLFix(BaseSpecHook):
    """SpecHook for transforming an old python package URL to a new URL
    (e.g. https://pypi.python.org/packages/* to
    https://files.pythonhosted.org/packages/* or https://pypi.python.org/pypi/*
    to https://pypi.org/project/*).

    """

    CATEGORIES: PackageCategories = [PackageCategory.python]

    URL_TRANSFORMATIONS: List[Tuple[str, str]] = [
        (r'https?://pypi\.python\.org/pypi/', 'https://pypi.org/project/'),
    ]
    SOURCES_URL_TRANSFORMATIONS: List[Tuple[str, str]] = [
        (r'https?://pypi(\.python)?\.org/', 'https://files.pythonhosted.org/'),
    ]

    @classmethod
    def run(cls, spec_file, rebase_spec_file, **kwargs):
        for tag in rebase_spec_file.tags.filter(name='URL'):
            value = rebase_spec_file.get_raw_tag_value(tag.name)
            if value:
                rebase_spec_file.set_raw_tag_value(tag.name, cls._transform_url(value))
        for tag in rebase_spec_file.tags.filter(name='Source*'):
            value = rebase_spec_file.get_raw_tag_value(tag.name)
            if value:
                rebase_spec_file.set_raw_tag_value(tag.name, cls._transform_sources_url(value))
        rebase_spec_file.save()

    @classmethod
    def _transform_url(cls, value):
        for trans in cls.URL_TRANSFORMATIONS:
            value = re.sub(trans[0], trans[1], value)
        return value

    @classmethod
    def _transform_sources_url(cls, value):
        for trans in cls.SOURCES_URL_TRANSFORMATIONS:
            value = re.sub(trans[0], trans[1], value)
        return value



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import pytest  # type: ignore

from pkg_resources import parse_version

from rebasehelper.plugins.plugin_manager import plugin_manager
from rebasehelper.plugins.versioneers.anitya import Anitya
from rebasehelper.plugins.versioneers.pypi import PyPI
from rebasehelper.plugins.versioneers.npmjs import NPMJS
from rebasehelper.plugins.versioneers.cpan import CPAN
from rebasehelper.plugins.versioneers.hackage import Hackage


class TestVersioneer:

    @pytest.mark.parametrize('package, min_version', [
        ('vim-go', 'v1.13'),
        ('libtiff', '4.0.8'),
    ], ids=[
        'vim-go>=v1.13',
        'libtiff>=4.0.8',
    ])
    @pytest.mark.integration
    def test_anitya_versioneer(self, package, min_version):
        assert Anitya.name in plugin_manager.versioneers.plugins
        Anitya.API_URL = 'https://integration:4430/versioneers'
        version = plugin_manager.versioneers.run(Anitya.name, package, None)
        assert parse_version(version) >= parse_version(min_version)

    @pytest.mark.parametrize('package, min_version', [
        ('python-m2r', '0.1.7'),
        ('pyodbc', '4.0.17'),
    ], ids=[
        'python-m2r>=0.1.7',
        'pyodbc>=4.0.17',
    ])
    @pytest.mark.integration
    def test_pypi_versioneer(self, package, min_version):
        assert PyPI.name in plugin_manager.versioneers.plugins
        PyPI.API_URL = 'https://integration:4430/versioneers'
        version = plugin_manager.versioneers.run(PyPI.name, package, None)
        assert parse_version(version) >= parse_version(min_version)

    @pytest.mark.parametrize('package, min_version', [
        ('nodejs-read-pkg', '1.1.0'),
        ('uglify-js', '1.2.1')
    ], ids=[
        'nodejs-read-pkg>=1.1.0',
        'uglify-js>=1.2.1'
    ])
    @pytest.mark.integration
    def test_npmjs_versioneer(self, package, min_version):
        assert NPMJS.name in plugin_manager.versioneers.plugins
        NPMJS.API_URL = 'https://integration:4430/versioneers'
        version = plugin_manager.versioneers.run(NPMJS.name, package, None)
        assert parse_version(version) >= parse_version(min_version)

    @pytest.mark.parametrize('package, min_version', [
        ('perl-Task-Kensho-Toolchain', '0.39'),
        ('perl-Data-GUID', '0.049')
    ], ids=[
        'perl-Task-Kensho-Toolchain>=0.39',
        'perl-Data-GUID>=0.049'
    ])
    @pytest.mark.integration
    def test_cpan_versioneer(self, package, min_version):
        assert CPAN.name in plugin_manager.versioneers.plugins
        CPAN.API_URL = 'https://integration:4430/versioneers'
        version = plugin_manager.versioneers.run(CPAN.name, package, None)
        assert parse_version(version) >= parse_version(min_version)

    @pytest.mark.parametrize('package, min_version', [
        ('ghc-clock', '0.2.0.0'),
        ('cab', '0.2.17')
    ], ids=[
        'ghc-clock>=0.2.0.0',
        'cab>=0.2.17'
    ])
    @pytest.mark.integration
    def test_hackage_versioneer(self, package, min_version):
        assert Hackage.name in plugin_manager.versioneers.plugins
        Hackage.API_URL = 'https://integration:4430/versioneers'
        version = plugin_manager.versioneers.run(Hackage.name, package, None)
        assert parse_version(version) >= parse_version(min_version)
